""" Functions for estimating the quality of spike sorting results. These
functions estimate false positive and false negative fractions.
"""

from __future__ import division

import scipy as sp
from scipy.spatial.distance import cdist
import quantities as pq
import neo

from progress_indicator import ProgressIndicator
from . import SpykeException
from conversions import spikes_to_spike_train


def get_refperiod_violations(spike_trains, refperiod, progress=None):
    """ Return the refractory period violations in the given spike trains
    for the specified refractory period.

    :param dict spike_trains: Dictionary of lists of
        :class:`neo.core.SpikeTrain` objects.
    :param refperiod: The refractory period (time).
    :type refperiod: Quantity scalar
    :param progress: Set this parameter to report progress.
    :type progress: :class:`.progress_indicator.ProgressIndicator`
    :returns: Two values:

        * The total number of violations.
        * A dictionary (with the same indices as ``spike_trains``) of
          arrays with violation times (Quantity 1D with the same unit as
          ``refperiod``) for each spike train.
    :rtype: int, dict """
    if type(refperiod) != pq.Quantity or \
            refperiod.simplified.dimensionality != pq.s.dimensionality:
        raise ValueError('refperiod must be a time quantity!')

    if not progress:
        progress = ProgressIndicator()

    total_violations = 0
    violations = {}
    for u, tL in spike_trains.iteritems():
        violations[u] = []
        for i, t in enumerate(tL):
            st = t.copy()
            st.sort()
            isi = sp.diff(st)

            violations[u].append(st[isi < refperiod].rescale(refperiod.units))
            total_violations += len(violations[u][i])

            progress.step()

    return total_violations, violations


def calculate_refperiod_fp(num_spikes, refperiod, violations, total_time):
    """ Return the rate of false positives calculated from refractory period
    calculations for each unit. The equation used is described in
    (Hill et al. The Journal of Neuroscience. 2011).

    :param dict num_spikes: Dictionary of total number of spikes,
        indexed by unit.
    :param refperiod: The refractory period (time). If the spike sorting
        algorithm includes a censored period (a time after a spike during
        which no new spikes can be found), subtract it from the refractory
        period before passing it to this function.
    :type refperiod: Quantity scalar
    :param dict violations: Dictionary of total number of violations,
        indexed the same as num_spikes.
    :param total_time: The total time in which violations could have occured.
    :type total_time: Quantity scalar

    :returns: A dictionary of false positive rates indexed by unit.
        Note that values above 0.5 can not be directly interpreted as a
        false positive rate! These very high values can e.g. indicate
        that the generating processes are not independent.
    """
    if type(refperiod) != pq.Quantity or \
            refperiod.simplified.dimensionality != pq.s.dimensionality:
        raise ValueError('refperiod must be a time quantity!')

    fp = {}
    factor = total_time / (2 * refperiod)
    for u, n in num_spikes.iteritems():
        if n == 0:
            fp[u] = 0
            continue
        zw = (violations[u] * factor / n ** 2).simplified

        if zw > 0.25:
            fp[u] = 0.5 + sp.sqrt(0.25 - zw).imag
            continue
        fp[u] = 0.5 - sp.sqrt(0.25 - zw)

    return fp


def _multi_norm(x, mean):
    """ Evaluate pdf of multivariate normal distribution with a mean
        at rows of x with high precision.
    """
    d = x.shape[1]
    fac = (2 * sp.pi) ** (-d / 2.0)
    y = cdist(x, sp.atleast_2d(mean), 'sqeuclidean') * -0.5
    return fac * sp.exp(sp.longdouble(y))


def _fast_overlap_whitened(spike_arrays, means):
    units = spike_arrays.keys()
    spikes = {u: spike_arrays[u].shape[1] for u in spike_arrays.iterkeys()}

    prior = {}
    total_spikes = 0
    for u, mean in means.iteritems():
        total_spikes += spikes[u]
    if total_spikes < 1:
        return {u: (0.0, 0.0) for u in units}, {}

    # Arrays of unnormalized posteriors (likelihood times prior)
    # for all units
    posterior = {}

    false_positive = {}
    false_negative = {}
    for u in units:
        prior[u] = spikes[u] / total_spikes
        false_positive[u] = 0
        false_negative[u] = 0

    # Calculate posteriors
    for u1 in units[:]:
        if not spikes[u1]:
            units.remove(u1)
            continue
        posterior[u1] = {}
        for u2, mean in means.iteritems():
            llh = _multi_norm(spike_arrays[u1].T, mean)
            posterior[u1][u2] = llh * prior[u2]

    # Calculate pairwise false positives/negatives
    singles = {u: {} for u in units}
    for i, u1 in enumerate(units):
        u1 = units[i]
        for u2 in units[i + 1:]:
            f1 = sp.sum(posterior[u1][u2] /
                        (posterior[u1][u1] + posterior[u1][u2]),
                        dtype=sp.double)

            f2 = sp.sum(posterior[u2][u1] /
                        (posterior[u2][u1] + posterior[u2][u2]),
                        dtype=sp.double)

            singles[u1][u2] = (f1 / spikes[u1] if spikes[u1] else 0,
                               f2 / spikes[u1] if spikes[u1] else 0)
            singles[u2][u1] = (f2 / spikes[u2] if spikes[u2] else 0,
                               f1 / spikes[u2] if spikes[u2] else 0)

    # Calculate complete false positives/negatives with extended bayes
    for u1 in units:
        numerator = posterior[u1][u1]
        normalizer = sum(posterior[u1][u2] for u2 in units)
        false_positive[u1] = sp.sum((normalizer - numerator) / normalizer)

        other_units = units[:]
        other_units.remove(u1)
        numerator = sp.vstack((posterior[u][u1] for u in other_units))
        normalizer = sp.vstack(sum(posterior[u][u2] for u2 in units) for u in other_units)
        false_negative[u1] = sp.sum(numerator / normalizer)

    # Prepare return values, convert sums to means
    totals = {}
    for u, fp in false_positive.iteritems():
        fn = false_negative[u]
        if not spikes[u]:
            totals[u] = (0, 0)
        else:
            num = spikes[u]
            totals[u] = (fp / num, fn / num)
    return totals, singles


def _pair_overlap(waves1, waves2, mean1, mean2, cov1, cov2):
    """ Calculate FP/FN estimates for two gaussian clusters
    """
    from sklearn import mixture

    means = sp.vstack([[mean1], [mean2]])
    covars = sp.vstack([[cov1], [cov2]])
    weights = sp.array([waves1.shape[1], waves2.shape[1]], dtype=float)
    weights /= weights.sum()

    # Create mixture of two Gaussians from the existing estimates
    mix = mixture.GMM(n_components=2, covariance_type='full', init_params='')
    mix.covars_ = covars
    mix.weights_ = weights
    mix.means_ = means

    posterior1 = mix.predict_proba(waves1.T)[:, 1]
    posterior2 = mix.predict_proba(waves2.T)[:, 0]

    return (posterior1.mean(), posterior2.sum() / len(posterior1),
            posterior2.mean(), posterior1.sum() / len(posterior2))


def _object_has_size(obj, size):
    """ Return if the object, which could be either a neo.Spike or ndarray,
    has the given size. """
    if isinstance(obj, neo.Spike):
        return obj.waveform.size == size
    return obj.size == size


def overlap_fp_fn(spikes, means=None, covariances=None):
    """ Return dicts of tuples (False positive rate, false negative rate)
    indexed by unit. This function needs :mod:`sklearn` if
    ``covariances`` is not set to ``'white'``.

    This function estimates the pairwise and total false positive and false
    negative rates for a number of waveform clusters. The results can be
    interpreted as follows: False positives are the fraction of spikes in a
    cluster that is estimated to belong to a different cluster (a specific
    cluster for pairwise results or any other cluster for total results).
    False negatives are the number spikes from other clusters that are
    estimated to belong to a given cluster (also expressed as fraction, this
    number can be larger than 1 in extreme cases).

    Details for the calculation can be found in
    (Hill et al. The Journal of Neuroscience. 2011).
    The calculation for total false positive and false negative rates does
    not follow Hill et al., who propose a simple addition of pairwise
    probabilities. Instead, the total error probabilities are estimated
    using all clusters at once.

    :param dict spikes: Dictionary, indexed by unit, of lists of
        spike waveforms as :class:`neo.core.Spike` objects or numpy arrays.
        If the waveforms have multiple channels, they will be flattened
        automatically. All waveforms need to have the same number of samples.
    :param dict means: Dictionary, indexed by unit, of lists of
        spike waveforms as :class:`neo.core.Spike` objects or numpy arrays.
        Means for units that are not in this dictionary will be estimated
        using the spikes. Note that if you pass ``'white'`` for
        ``covariances`` and you want to provide means, they have to be
        whitened in the same way as the spikes.
        Default: None, means will be estimated from data.
    :param covariances: Dictionary, indexed by unit, of lists of
        covariance matrices. Covariances  for units that are not in this
        dictionary will be estimated using the spikes. It is useful to give
        a covariance matrix if few spikes are present - consider using the
        noise covariance. If you use prewhitened spikes (i.e. all clusters
        are normal distributed, so their covariance matrix is the identity),
        you can pass ``'white'`` here. The calculation will be much faster in
        this case and the sklearn package is not required.
        Default: None, covariances will estimated from data.
    :type covariances: dict or str
    :returns: Two values:

        * A dictionary (indexed by unit) of total
          (false positive rate, false negative rate) tuples.
        * A dictionary of dictionaries, both indexed by units,
          of pairwise (false positive rate, false negative rate) tuples.
    :rtype: dict, dict
    """
    units = spikes.keys()

    total_spikes = 0
    for spks in spikes.itervalues():
        total_spikes += len(spks)
    if total_spikes < 1:
        return {u: (0.0, 0.0) for u in units}, {}

    if means is None:
        means = {}
    white = False
    if covariances is None:
        covariances = {}
    elif covariances == 'white':
        white = True
        covariances = {}

    # Convert Spike objects to arrays
    dimensionality = None
    spike_arrays = {}
    for u, spks in spikes.iteritems():
        spikelist = []
        if not spks or (len(spks) < 2 and u not in covariances):
            units.remove(u)
            continue
        for s in spks:
            if isinstance(s, neo.Spike):
                spikelist.append(
                    sp.asarray(s.waveform.rescale(pq.uV)).T.flatten())
            else:
                spikelist.append(s)
        spike_arrays[u] = sp.array(spikelist).T
        if dimensionality is None:
            dimensionality = spike_arrays[u].shape[0]
        elif dimensionality != spike_arrays[u].shape[0]:
            raise SpykeException('All spikes need to have the same number'
                                 'of samples!')

    if not units:
        return {}, {}
    if len(units) == 1:
        return {units[0]: (0.0, 0.0)}, {}

    # Convert or calculate means and covariances
    shaped_means = {}
    covs = {}
    if white:
        cov = sp.eye(dimensionality)
        covariances = {u: cov for u in units}

    for u in units:
        if u in means and _object_has_size(means[u], dimensionality):
            mean = means[u]
            if isinstance(mean, neo.Spike):
                shaped_means[u] = sp.asarray(
                    mean.waveform.rescale(pq.uV)).T.flatten()
            else:
                shaped_means[u] = means[u].T.flatten()
        else:
            shaped_means[u] = spike_arrays[u].mean(axis=1)

    if white:
        return _fast_overlap_whitened(spike_arrays, shaped_means)

    for u in units:
        if u not in covariances:
            covs[u] = sp.cov(spike_arrays[u])
        else:
            covs[u] = covariances[u]

    # Calculate pairwise false positives/negatives
    singles = {u: {} for u in units}
    for i, u1 in enumerate(units):
        u1 = units[i]
        for u2 in units[i + 1:]:
            error_rates = _pair_overlap(
                spike_arrays[u1], spike_arrays[u2],
                shaped_means[u1], shaped_means[u2],
                covs[u1], covs[u2])
            singles[u1][u2] = error_rates[0:2]
            singles[u2][u1] = error_rates[2:4]

    # Calculate complete false positives/negatives
    import sklearn
    mix = sklearn.mixture.GMM(n_components=2, covariance_type='full')
    mix_means = []
    mix_covars = []
    mix_weights = []
    for u in units:
        mix_means.append(shaped_means[u])
        mix_covars.append([covs[u]])
        mix_weights.append(spike_arrays[u].shape[1])
    mix.means_ = sp.vstack(mix_means)
    mix.covars_ = sp.vstack(mix_covars)
    mix_weights = sp.array(mix_weights, dtype=float)
    mix_weights /= mix_weights.sum()
    mix.weights_ = mix_weights

    # P(spikes of unit[i] in correct cluster)
    post_mean = sp.zeros(len(units))

    # sum(P(spikes of unit[i] in cluster[j])
    post_sum = sp.zeros((len(units), len(units)))

    for i, u in enumerate(units):
        posterior = mix.predict_proba(spike_arrays[u].T)
        post_mean[i] = posterior[:, i].mean()
        post_sum[i, :] = posterior.sum(axis=0)

    totals = {}
    for i, u in enumerate(units):
        fp = 1.0 - post_mean[i]
        ind = range(len(units))
        ind.remove(i)
        fn = post_sum[ind, i].sum() / float(spike_arrays[u].shape[1])
        totals[u] = (fp, fn)

    return totals, singles


def variance_explained(spikes, means=None, noise=None):
    """ Returns the fraction of variance in each channel that is explained
    by the means.

    Values below 0 or above 1 for large data sizes indicate
    that some assumptions were incorrect (e.g. about channel noise) and
    the results should not be trusted.

    :param dict spikes: Dictionary, indexed by unit, of
        :class:`neo.core.SpikeTrain` objects (where the ``waveforms``
        member includes the spike waveforms) or lists of
        :class:`neo.core.Spike` objects.
    :param dict means: Dictionary, indexed by unit, of lists of
        spike waveforms as :class:`neo.core.Spike` objects or numpy arrays.
        Means for units that are not in this dictionary will be estimated
        using the spikes.
        Default: None - means will be estimated from given spikes.
    :type noise: Quantity 1D
    :param noise: The known noise levels (as variance) per channel of the
        original data. This should be estimated from the signal periods
        that do not contain spikes, otherwise the explained variance
        could be overestimated. If None, the estimate of explained variance
        is done without regard for noise.
        Default: None
    :return dict: A dictionary of arrays, both indexed by unit. If ``noise``
        is ``None``, the  dictionary contains
        the fraction of explained variance per channel without taking noise
        into account. If ``noise`` is given, it contains the fraction of
        variance per channel explained by the means and given noise level
        together.
    """
    ret = {}
    if means is None:
        means = {}
    for u, spks in spikes.iteritems():
        train = spks
        if not isinstance(train, neo.SpikeTrain):
            train = spikes_to_spike_train(spks)
        if u in means and means[u].waveform.shape[0] == train.waveforms.shape[1]:
            spike = means[u]
        else:
            spike = neo.Spike(0)
            spike.waveform = sp.mean(train.waveforms, axis=0)

        orig = sp.mean(sp.var(train.waveforms, axis=1), axis=0)
        waves = train.waveforms - spike.waveform
        new = sp.mean(sp.var(waves, axis=1), axis=0)

        if noise is not None:
            ret[u] = sp.asarray(1 - (new - noise) / orig)
        else:
            ret[u] = sp.asarray(1 - new / orig)

    return ret



try:
    import unittest2 as ut
    assert ut  # Suppress pyflakes warning about redefinition of unused ut
except ImportError:
    import unittest as ut


from numpy.testing import assert_array_equal, assert_array_almost_equal
import scipy as sp
import quantities as pq
from spykeutils.monkeypatch import quantities_patch

assert quantities_patch  # Suppress pyflakes warning, patch applied by loading


class TestQuantityMax(ut.TestCase):
    def test_returns_global_max(self):
        a = sp.array([[1, 2], [3, 4]]) * pq.s
        self.assertEqual(4, a.max())

    def test_returns_axiswise_max(self):
        a = sp.array([[1, 2], [3, 4]]) * pq.s
        axis = 1
        expected = sp.array([2, 4]) * pq.s
        assert_array_equal(expected, a.max(axis=axis))

    def test_returns_result_in_out_array(self):
        a = sp.array([[1, 2], [3, 4]]) * pq.s
        out = 0 * pq.s
        expected = sp.array([4])
        a.max(out=out)
        self.assertEqual(expected, out)


class TestQuantityMin(ut.TestCase):
    def test_returns_global_min(self):
        a = sp.array([[1, 2], [3, 4]]) * pq.s
        self.assertEqual(1, a.min())

    def test_returns_axiswise_min(self):
        a = sp.array([[1, 2], [3, 4]]) * pq.s
        axis = 1
        expected = sp.array([1, 3]) * pq.s
        assert_array_equal(expected, a.min(axis=axis))

    def test_returns_result_in_out_array(self):
        a = sp.array([[1, 2], [3, 4]]) * pq.s
        out = 0 * pq.s
        expected = 1 * pq.s
        a.min(out=out)
        self.assertEqual(expected, out)


if __name__ == '__main__':
    ut.main()



""" This package provides support for writing plugins for Spyke Viewer.
It belongs to `spykeutils` so that plugins can be executed in an evironment
where the `spykeviewer` package and its dependencies are not installed
(e.g. servers).

`spykeutils` installs a script named "spykeplugin" that can be used to start
plugins directly from the command line, supplying selection and plugin
parameter information. It is also the default script that Spyke Viewer uses
when starting plugins remotely. If you want to implement your own script for
starting plugins remotely, e.g. on a server, you should conform to the
interface of this script.

:mod:`analysis_plugin` Module
-----------------------------

.. automodule:: spykeutils.plugin.analysis_plugin
    :members:
    :show-inheritance:

:mod:`data_provider` Module
---------------------------

.. automodule:: spykeutils.plugin.data_provider
    :members:
    :show-inheritance:

:mod:`gui_data` Module
----------------------

.. automodule:: spykeutils.plugin.gui_data
"""


try:
    import unittest2 as ut
    assert ut  # Suppress pyflakes warning about redefinition of unused ut
except ImportError:
    import unittest as ut

from builders import arange_spikes
from numpy.testing import assert_array_equal, assert_array_almost_equal
from spykeutils import tools
import neo
import neo.io.tools
import neo.test.tools
import quantities as pq
import scipy as sp


class TestApplyToDict(ut.TestCase):
    @staticmethod
    def fn(train, multiplier=1):
        return multiplier * train.size

    def test_maps_function_to_each_spike_train(self):
        st_dict = {'a': [arange_spikes(5 * pq.s), arange_spikes(4 * pq.s)],
                   'b': [arange_spikes(7 * pq.s)]}
        expected = {'a': [4, 3], 'b': [6]}
        actual = tools.apply_to_dict(self.fn, st_dict)
        self.assertEqual(expected, actual)

    def test_works_on_empty_lists(self):
        st_dict = {'a': [], 'b': []}
        expected = {'a': [], 'b': []}
        actual = tools.apply_to_dict(self.fn, st_dict)
        self.assertEqual(expected, actual)

    def test_works_on_empty_dict(self):
        st_dict = {}
        expected = {}
        actual = tools.apply_to_dict(self.fn, st_dict)
        self.assertEqual(expected, actual)

    def test_allows_to_pass_additional_args(self):
        st_dict = {'a': [arange_spikes(5 * pq.s), arange_spikes(4 * pq.s)],
                   'b': [arange_spikes(7 * pq.s)]}
        expected = {'a': [8, 6], 'b': [12]}
        actual = tools.apply_to_dict(self.fn, st_dict, 2)
        self.assertEqual(expected, actual)


class TestBinSpikeTrains(ut.TestCase):
    def test_bins_spike_train_using_its_properties(self):
        a = neo.SpikeTrain(
            sp.array([1000.0]) * pq.ms, t_start=500.0 * pq.ms,
            t_stop=1500.0 * pq.ms)
        sampling_rate = 4.0 * pq.Hz
        expected = {0: [sp.array([0, 0, 1, 0])]}
        expectedBins = sp.array([0.5, 0.75, 1.0, 1.25, 1.5]) * pq.s
        actual, actualBins = tools.bin_spike_trains({0: [a]}, sampling_rate)
        self.assertEqual(len(expected), len(actual))
        self.assertEqual(len(expected[0]), len(actual[0]))
        assert_array_equal(expected[0][0], actual[0][0])
        assert_array_almost_equal(
            expectedBins, actualBins.rescale(expectedBins.units))

    def test_bins_spike_train_using_passed_properties(self):
        a = neo.SpikeTrain(
            sp.array([1.0]) * pq.s, t_start=0.0 * pq.s, t_stop=5.0 * pq.s)
        sampling_rate = 4.0 * pq.Hz
        t_start = 0.5 * pq.s
        t_stop = 1.5 * pq.s
        expected = {0: [sp.array([0, 0, 1, 0])]}
        expectedBins = sp.array([0.5, 0.75, 1.0, 1.25, 1.5]) * pq.s
        actual, actualBins = tools.bin_spike_trains(
            {0: [a]}, sampling_rate=sampling_rate, t_start=t_start,
            t_stop=t_stop)
        self.assertEqual(len(expected), len(actual))
        self.assertEqual(len(expected[0]), len(actual[0]))
        assert_array_equal(expected[0][0], actual[0][0])
        assert_array_almost_equal(
            expectedBins, actualBins.rescale(expectedBins.units))

    def test_uses_max_spike_train_interval(self):
        a = arange_spikes(5 * pq.s)
        b = arange_spikes(7 * pq.s, 15 * pq.s)
        sampling_rate = 4.0 * pq.Hz
        expectedBins = sp.arange(0.0, 15.1, 0.25) * pq.s
        actual, actualBins = tools.bin_spike_trains(
            {0: [a, b]}, sampling_rate=sampling_rate)
        assert_array_almost_equal(
            expectedBins, actualBins.rescale(expectedBins.units))

    def test_handles_bin_size_which_is_not_divisor_of_duration(self):
        a = arange_spikes(5 * pq.s)
        sampling_rate = 1.0 / 1.3 * pq.Hz
        expected = {0: [sp.array([1, 1, 1, 1])]}
        expectedBins = sp.array([0.0, 1.3, 2.6, 3.9, 5.2]) * pq.s
        actual, actualBins = tools.bin_spike_trains({0: [a]}, sampling_rate)
        self.assertEqual(len(expected), len(actual))
        self.assertEqual(len(expected[0]), len(actual[0]))
        assert_array_equal(expected[0][0], actual[0][0])
        assert_array_almost_equal(
            expectedBins, actualBins.rescale(expectedBins.units))


class TestConcatenateSpikeTrains(ut.TestCase):
    def test_concatenates_spike_trains(self):
        a = arange_spikes(3.0 * pq.s)
        b = arange_spikes(2.0 * pq.s, 5.0 * pq.s)
        expected = arange_spikes(5.0 * pq.s)
        actual = tools.concatenate_spike_trains((a, b))
        assert_array_almost_equal(expected, actual)

    def test_t_start_is_min_of_all_trains(self):
        a = arange_spikes(3.0 * pq.s, 5.0 * pq.s)
        b = arange_spikes(1.0 * pq.s, 6.0 * pq.s)
        expected = 1.0 * pq.s
        actual = tools.concatenate_spike_trains((a, b)).t_start
        self.assertAlmostEqual(expected, actual)

    def test_t_stop_is_max_of_all_trains(self):
        a = arange_spikes(3.0 * pq.s, 5.0 * pq.s)
        b = arange_spikes(1.0 * pq.s, 6.0 * pq.s)
        expected = 6.0 * pq.s
        actual = tools.concatenate_spike_trains((a, b)).t_stop
        self.assertAlmostEqual(expected, actual)


class TestRemoveFromHierarchy(ut.TestCase):
    SEGMENTS = 5
    CHANNEL_GROUPS = 4
    UNITS = 3
    CHANNELS = 4

    @classmethod
    def create_hierarchy(cls, many_to_many):
        b = neo.Block()

        for ns in range(cls.SEGMENTS):
            b.segments.append(neo.Segment())

        channels = []
        if many_to_many:
            channels = [neo.RecordingChannel(name='Shared %d' % i,
                                             index=i + cls.CHANNELS)
                        for i in range(cls.CHANNELS / 2)]

        for ng in range(cls.CHANNEL_GROUPS):
            rcg = neo.RecordingChannelGroup()
            for nu in range(cls.UNITS):
                unit = neo.Unit()
                for ns in range(cls.SEGMENTS):
                    spike = neo.Spike(0 * pq.s)
                    unit.spikes.append(spike)
                    b.segments[ns].spikes.append(spike)

                    st = neo.SpikeTrain([] * pq.s, 0 * pq.s)
                    unit.spiketrains.append(st)
                    b.segments[ns].spiketrains.append(st)

                rcg.units.append(unit)

            if not many_to_many:
                for nc in range(cls.CHANNELS):
                    rc = neo.RecordingChannel(
                        name='Single %d' % nc, index=nc)
                    rc.recordingchannelgroups.append(rcg)
                    rcg.recordingchannels.append(rc)
            else:
                for nc in range(cls.CHANNELS):
                    if nc % 2 == 0:
                        rc = neo.RecordingChannel(
                            name='Single %d' % (nc / 2), index=nc / 2)
                    else:
                        rc = channels[nc / 2]
                    rc.recordingchannelgroups.append(rcg)
                    rcg.recordingchannels.append(rc)
            rcg.channel_indexes = sp.array(
                [c.index for c in rcg.recordingchannels])
            rcg.channel_names = sp.array(
                [c.name for c in rcg.recordingchannels])

            b.recordingchannelgroups.append(rcg)

        neo.io.tools.create_many_to_one_relationship(b)
        return b

    def test_remove_block(self):
        block = self.create_hierarchy(False)
        comp = self.create_hierarchy(False)
        tools.remove_from_hierarchy(block)

        neo.test.tools.assert_same_sub_schema(block, comp)

    def test_remove_segment_no_orphans(self):
        block = self.create_hierarchy(False)
        comp = self.create_hierarchy(False)

        seg = block.segments[1]
        tools.remove_from_hierarchy(seg)

        self.assertFalse(seg in block.segments)
        self.assertEqual(len(block.list_units),
                         self.UNITS * self.CHANNEL_GROUPS)
        for u in block.list_units:
            self.assertEqual(len(u.spikes), self.SEGMENTS - 1)
            self.assertEqual(len(u.spiketrains), self.SEGMENTS - 1)
        neo.test.tools.assert_same_sub_schema(seg, comp.segments[1])

    def test_remove_segment_keep_orphans(self):
        block = self.create_hierarchy(False)
        comp = self.create_hierarchy(False)

        seg = block.segments[1]
        tools.remove_from_hierarchy(seg, False)

        self.assertFalse(seg in block.segments)
        self.assertEqual(len(block.list_units),
                         self.UNITS * self.CHANNEL_GROUPS)
        for u in block.list_units:
            self.assertEqual(len(u.spikes), self.SEGMENTS)
            self.assertEqual(len(u.spiketrains), self.SEGMENTS)
        neo.test.tools.assert_same_sub_schema(seg, comp.segments[1])

    def test_remove_channel_group_no_orphans(self):
        block = self.create_hierarchy(False)
        comp = self.create_hierarchy(False)

        rcg = block.recordingchannelgroups[1]
        tools.remove_from_hierarchy(rcg)

        self.assertFalse(rcg in block.recordingchannelgroups)
        self.assertEqual(len(block.segments), self.SEGMENTS)
        for s in block.segments:
            self.assertEqual(len(s.spikes),
                             self.UNITS * (self.CHANNEL_GROUPS - 1))
            self.assertEqual(len(s.spiketrains),
                             self.UNITS * (self.CHANNEL_GROUPS - 1))
        neo.test.tools.assert_same_sub_schema(rcg,
                                              comp.recordingchannelgroups[1])

    def test_remove_channel_group_keep_orphans(self):
        block = self.create_hierarchy(False)
        comp = self.create_hierarchy(False)

        rcg = block.recordingchannelgroups[1]
        tools.remove_from_hierarchy(rcg, False)

        self.assertFalse(rcg in block.recordingchannelgroups)
        self.assertEqual(len(block.segments), self.SEGMENTS)
        for s in block.segments:
            self.assertEqual(len(s.spikes),
                             self.UNITS * self.CHANNEL_GROUPS)
            self.assertEqual(len(s.spiketrains),
                             self.UNITS * self.CHANNEL_GROUPS)
        neo.test.tools.assert_same_sub_schema(rcg,
                                              comp.recordingchannelgroups[1])

    def test_remove_channel(self):
        block = self.create_hierarchy(False)
        comp = self.create_hierarchy(False)

        rc = block.list_recordingchannels[5]
        tools.remove_from_hierarchy(rc)
        self.assertFalse(rc in block.list_recordingchannels)
        neo.test.tools.assert_same_sub_schema(rc,
                                              comp.list_recordingchannels[5])

        self.assertEqual(len(block.segments), self.SEGMENTS)
        self.assertEqual(len(block.recordingchannelgroups),
                         self.CHANNEL_GROUPS)
        self.assertEqual(len(block.list_recordingchannels),
                         self.CHANNEL_GROUPS * self.CHANNELS - 1)

        # Should be removed from its own channel group
        rcg = rc.recordingchannelgroups[0]
        self.assertEqual(len(rcg.recordingchannels), self.CHANNELS - 1)
        self.assertEqual(rcg.channel_indexes.shape[0], self.CHANNELS - 1)
        self.assertEqual(rcg.channel_names.shape[0], self.CHANNELS - 1)
        self.assertFalse(rc.index in rcg.channel_indexes)
        self.assertFalse(rc.name in rcg.channel_names)

    def test_remove_unique_channel_many_to_many(self):
        block = self.create_hierarchy(True)
        comp = self.create_hierarchy(True)
        self.assertEqual(
            len(block.list_recordingchannels),
            self.CHANNEL_GROUPS * (self.CHANNELS / 2) + (self.CHANNELS / 2))

        rc = block.list_recordingchannels[0]  # Unique channel
        tools.remove_from_hierarchy(rc)

        neo.test.tools.assert_same_sub_schema(rc,
                                              comp.list_recordingchannels[0])
        self.assertFalse(rc in block.list_recordingchannels)

        self.assertEqual(len(block.segments), self.SEGMENTS)
        self.assertEqual(len(block.recordingchannelgroups),
                         self.CHANNEL_GROUPS)
        self.assertEqual(
            len(block.list_recordingchannels),
            self.CHANNEL_GROUPS * (self.CHANNELS / 2) + (self.CHANNELS / 2) - 1)

        # Should be removed from its own channel group
        rcg = rc.recordingchannelgroups[0]
        self.assertEqual(len(rcg.recordingchannels), self.CHANNELS - 1)
        self.assertEqual(rcg.channel_indexes.shape[0], self.CHANNELS - 1)
        self.assertEqual(rcg.channel_names.shape[0], self.CHANNELS - 1)
        self.assertFalse(rc.index in rcg.channel_indexes)
        self.assertFalse(rc.name in rcg.channel_names)

    def test_remove_shared_channel_many_to_many(self):
        block = self.create_hierarchy(True)
        comp = self.create_hierarchy(True)
        self.assertEqual(
            len(block.list_recordingchannels),
            self.CHANNEL_GROUPS * (self.CHANNELS / 2) + (self.CHANNELS / 2))

        rc = block.list_recordingchannels[1]  # Shared channel
        tools.remove_from_hierarchy(rc)

        neo.test.tools.assert_same_sub_schema(rc,
                                              comp.list_recordingchannels[1])
        self.assertFalse(rc in block.list_recordingchannels)

        self.assertEqual(len(block.segments), self.SEGMENTS)
        self.assertEqual(len(block.recordingchannelgroups),
                         self.CHANNEL_GROUPS)
        self.assertEqual(
            len(block.list_recordingchannels),
            self.CHANNEL_GROUPS * (self.CHANNELS / 2) + (self.CHANNELS / 2) - 1)

        # Should be removed from all channel groups
        for rcg in block.recordingchannelgroups:
            self.assertEqual(len(rcg.recordingchannels), self.CHANNELS - 1)
            self.assertEqual(rcg.channel_indexes.shape[0], self.CHANNELS - 1)
            self.assertEqual(rcg.channel_names.shape[0], self.CHANNELS - 1)
            self.assertFalse(rc.index in rcg.channel_indexes)
            self.assertFalse(rc.name in rcg.channel_names)

    def test_remove_unit_no_orphans(self):
        block = self.create_hierarchy(False)
        comp = self.create_hierarchy(False)

        unit = block.list_units[5]
        tools.remove_from_hierarchy(unit)

        self.assertFalse(unit in block.list_units)
        self.assertEqual(len(block.list_units),
                         self.UNITS * self.CHANNEL_GROUPS - 1)
        self.assertEqual(len(block.segments), self.SEGMENTS)
        self.assertEqual(len(block.recordingchannelgroups),
                         self.CHANNEL_GROUPS)
        for seg in block.segments:
            self.assertEqual(len(seg.spikes),
                             self.UNITS * self.CHANNEL_GROUPS - 1)
            self.assertEqual(len(seg.spiketrains),
                             self.UNITS * self.CHANNEL_GROUPS - 1)
            self.assertFalse(unit in [s.unit for s in seg.spikes])
            self.assertFalse(unit in [st.unit for st in seg.spiketrains])
        neo.test.tools.assert_same_sub_schema(unit, comp.list_units[5])

    def test_remove_unit_keep_orphans(self):
        block = self.create_hierarchy(False)
        comp = self.create_hierarchy(False)

        unit = block.list_units[5]
        tools.remove_from_hierarchy(unit, False)

        self.assertFalse(unit in block.list_units)
        self.assertEqual(len(block.list_units),
                         self.UNITS * self.CHANNEL_GROUPS - 1)
        self.assertEqual(len(block.segments), self.SEGMENTS)
        self.assertEqual(len(block.recordingchannelgroups),
                         self.CHANNEL_GROUPS)
        for seg in block.segments:
            self.assertEqual(len(seg.spikes),
                             self.UNITS * self.CHANNEL_GROUPS)
            self.assertEqual(len(seg.spiketrains),
                             self.UNITS * self.CHANNEL_GROUPS)
            self.assertFalse(unit in [s.unit for s in seg.spikes])
            self.assertFalse(unit in [st.unit for st in seg.spiketrains])
        neo.test.tools.assert_same_sub_schema(unit, comp.list_units[5])

    def test_remove_spike(self):
        unit = neo.Unit()
        segment = neo.Segment()

        s = neo.Spike(0 * pq.s)
        unit.spikes.append(s)
        segment.spikes.append(s)
        s.unit = unit
        s.segment = segment

        st = neo.SpikeTrain([] * pq.s, 0 * pq.s)
        unit.spiketrains.append(st)
        segment.spiketrains.append(st)
        st.unit = unit
        st.segment = segment

        tools.remove_from_hierarchy(s)
        self.assertTrue(st in unit.spiketrains)
        self.assertTrue(st in segment.spiketrains)
        self.assertFalse(s in unit.spikes)
        self.assertFalse(s in segment.spikes)

    def test_remove_spiketrain(self):
        unit = neo.Unit()
        segment = neo.Segment()

        s = neo.Spike(0 * pq.s)
        unit.spikes.append(s)
        segment.spikes.append(s)
        s.unit = unit
        s.segment = segment

        st = neo.SpikeTrain([] * pq.s, 0 * pq.s)
        unit.spiketrains.append(st)
        segment.spiketrains.append(st)
        st.unit = unit
        st.segment = segment

        tools.remove_from_hierarchy(st)
        self.assertTrue(s in unit.spikes)
        self.assertTrue(s in segment.spikes)
        self.assertFalse(st in unit.spiketrains)
        self.assertFalse(st in segment.spiketrains)

    def test_extract_spikes(self):
        s1 = sp.zeros(10000)
        s2 = sp.ones(10000)
        t = sp.arange(0.0, 10.1, 1.0)

        sig1 = neo.AnalogSignal(s1 * pq.uV, sampling_rate=pq.kHz)
        sig2 = neo.AnalogSignal(s2 * pq.uV, sampling_rate=pq.kHz)
        train = neo.SpikeTrain(t * pq.s, 10 * pq.s)

        spikes = tools.extract_spikes(
            train, [sig1, sig2], 100 * pq.ms, 10 * pq.ms)

        self.assertEqual(len(spikes), 9)
        for s in spikes:
            self.assertAlmostEqual(s.waveform[:, 0].mean(), 0.0)
            self.assertAlmostEqual(s.waveform[:, 1].mean(), 1.0)

    def test_extract_different_spikes(self):
        s1 = sp.ones(10500)
        s2 = -sp.ones(10500)
        for i in xrange(10):
            s1[i * 1000 + 500:i * 1000 + 1500] *= i
            s2[i * 1000 + 500:i * 1000 + 1500] *= i
        t = sp.arange(0.0, 10.1, 1.0)

        sig1 = neo.AnalogSignal(s1 * pq.uV, sampling_rate=pq.kHz)
        sig2 = neo.AnalogSignal(s2 * pq.uV, sampling_rate=pq.kHz)
        train = neo.SpikeTrain(t * pq.s, 10 * pq.s)

        spikes = tools.extract_spikes(
            train, [sig1, sig2], 100 * pq.ms, 10 * pq.ms)

        self.assertEqual(len(spikes), 10)
        for i, s in enumerate(spikes):
            self.assertAlmostEqual(s.waveform[:, 0].mean(), i)
            self.assertAlmostEqual(s.waveform[:, 1].mean(), -i)

if __name__ == '__main__':
    ut.main()



import neo
import neo.description
import quantities as pq
import scipy as sp
import _scipy_quantities as spq


def apply_to_dict(fn, dictionary, *args):
    """ Applies a function to all spike trains in a dictionary of spike train
    sequences.

    :param function fn: Function to apply. Should take a
        :class:`neo.core.SpikeTrain` as first argument.
    :param dict dictionary: Dictionary of sequences of
        :class:`neo.core.SpikeTrain` objects to apply the function to.
    :param args: Additional arguments which will be passed to ``fn``.
    :returns: A new dictionary with the same keys as ``dictionary``.
    :rtype: dict
    """

    applied = {}
    for k in dictionary:
        applied[k] = [fn(st, *args) for st in dictionary[k]]
    return applied


def bin_spike_trains(trains, sampling_rate, t_start=None, t_stop=None):
    """ Creates binned representations of spike trains.

    :param dict trains: A dictionary of sequences of
        :class:`neo.core.SpikeTrain` objects.
    :param sampling_rate: The sampling rate which will be used to bin
        the spike trains as inverse time scalar.
    :type sampling_rate: Quantity scalar
    :type t_start: The desired time for the start of the first bin as time
        scalar. It will be the minimum start time of all spike trains if
        ``None`` is passed.
    :type t_start: Quantity scalar
    :param t_stop: The desired time for the end of the last bin as time scalar.
        It will be the maximum stop time of all spike trains if ``None`` is
        passed.
    :type t_stop: Quantity scalar
    :returns: A dictionary (with the same indices as ``trains``) of lists
        of spike train counts and the bin borders.
    :rtype: dict, Quantity 1D with time units
    """
    if t_start is None or t_stop is None:
        max_start, max_stop = maximum_spike_train_interval(trains)
        if t_start is None:
            t_start = max_start
        if t_stop is None:
            t_stop = max_stop

    t_start = t_start.rescale(t_stop.units)

    duration = t_stop - t_start
    num_bins = (sampling_rate * duration).simplified
    bins = sp.arange(num_bins + 1) * (duration / num_bins) + t_start
    return apply_to_dict(_bin_single_spike_train, trains, bins), bins


def _bin_single_spike_train(train, bins):
    """ Return a binned representation of SpikeTrain object.

    :param train: A spike train to bin.
    :type train: :class:`neo.core.SpikeTrain`
    :param bins: The bin edges, including the rightmost edge, with time units.
    :type bins: Quantity 1D
    :returns: The binned spike train.
    :rtype: 1-D array
    """
    return sp.histogram(train.rescale(bins.units), bins)[0]


def concatenate_spike_trains(trains):
    """ Concatenates spike trains.

    :param sequence trains: :class:`neo.core.SpikeTrain` objects to
        concatenate.
    :returns: A spike train consisting of the concatenated spike trains. The
        spikes will be in the order of the given spike trains and ``t_start``
        and ``t_stop`` will be set to the minimum and maximum value.
    :rtype: :class:`neo.core.SpikeTrain`
    """

    t_start, t_stop = maximum_spike_train_interval({0: trains})
    return neo.SpikeTrain(
        spq.concatenate([train.view(type=pq.Quantity) for train in trains]),
        t_start=t_start, t_stop=t_stop)


def minimum_spike_train_interval(
        trains, t_start=-sp.inf * pq.s, t_stop=sp.inf * pq.s):
    """ Computes the maximum starting time and minimum end time that all
    given spike trains share. This yields the shortest interval shared by all
    spike trains.

    :param dict trains: A dictionary of sequences of
        :class:`neo.core.SpikeTrain` objects.
    :param t_start: Minimal starting time to return.
    :type t_start: Quantity scalar
    :param t_stop: Maximum end time to return. If ``None``, infinity is used.
    :type t_stop: Quantity scalar
    :returns: Maximum shared t_start time and minimum shared t_stop time as
        time scalars.
    :rtype: Quantity scalar, Quantity scalar
    """
    if t_stop is None:
        t_stop = sp.inf * pq.s

    # Load data and find shortest spike train
    for st in trains.itervalues():
        if len(st) > 0:
            # Minimum length of spike of all spike trains for this unit
            t_start = max(t_start, max((t.t_start for t in st)))
            t_stop = min(t_stop, min((t.t_stop for t in st)))

    return t_start, t_stop


def maximum_spike_train_interval(
        trains, t_start=sp.inf * pq.s, t_stop=-sp.inf * pq.s):
    """ Computes the minimum starting time and maximum end time of all
    given spike trains. This yields an interval containing the spikes of
    all spike trains.

    :param dict trains: A dictionary of sequences of
        :class:`neo.core.SpikeTrain` objects.
    :param t_start: Maximum starting time to return.
    :type t_start: Quantity scalar
    :param t_stop: Minimum end time to return. If ``None``, infinity is used.
    :type t_stop: Quantity scalar
    :returns: Minimum t_start time and maximum t_stop time as time scalars.
    :rtype: Quantity scalar, Quantity scalar
    """
    if t_stop is None:
        t_stop = sp.inf * pq.s

    for st in trains.itervalues():
        if len(st) > 0:
            t_start = min(t_start, min((t.t_start for t in st)))
            t_stop = max(t_stop, max((t.t_stop for t in st)))

    return t_start, t_stop


def _handle_orphans(obj, remove):
    """ Removes half-orphaned Spikes and SpikeTrains that occur when
    removing an object upwards in the hierarchy.
    """
    if isinstance(obj, neo.Segment):
        for s in obj.spikes:
            if s.unit:
                if not remove:
                    s.segment = None
                else:
                    try:
                        s.unit.spikes.remove(s)
                    except ValueError:
                        pass

        for st in obj.spiketrains:
            if st.unit:
                if not remove:
                    st.segment = None
                else:
                    try:
                        st.unit.spiketrains.remove(st)
                    except ValueError:
                        pass
    elif isinstance(obj, neo.Unit):
        for s in obj.spikes:
            if s.segment:
                if not remove:
                    s.unit = None
                else:
                    try:
                        s.segment.spikes.remove(s)
                    except ValueError:
                        pass

        for st in obj.spiketrains:
            if st.segment:
                if not remove:
                    st.unit = None
                else:
                    try:
                        st.segment.spiketrains.remove(st)
                    except ValueError:
                        pass
    elif isinstance(obj, neo.RecordingChannelGroup):
        for u in obj.units:
            _handle_orphans(u, remove)


def remove_from_hierarchy(obj, remove_half_orphans=True):
    """ Removes a Neo object from the hierarchy it is embedded in. Mostly
    downward links are removed (except for possible links in
    :class:`neo.core.Spike` or :class:`neo.core.SpikeTrain` objects).
    For example, when ``obj`` is a :class:`neo.core.Segment`, the link from
    its parent :class:`neo.core.Block` will be severed. Also, all links to
    the segment from its spikes and spike trains will be severed.

    :param obj: The object to be removed.
    :type obj: Neo object
    :param bool remove_half_orphans: When True, :class:`neo.core.Spike`
        and :class:`neo.core.SpikeTrain` belonging to a
        :class:`neo.core.Segment` or :class:`neo.core.Unit` removed by
        this function will be removed from the hierarchy as well, even
        if they are still linked from a :class:`neo.core.Unit` or
        :class:`neo.core.Segment`, respectively. In this case, their
        links to the hierarchy defined by ``obj`` will be kept intact.
    """
    classname = type(obj).__name__

    # Parent for arbitrary object
    if classname in neo.description.many_to_one_relationship:
        for n in neo.description.many_to_one_relationship[classname]:
            p = getattr(obj, n.lower())
            if p is None:
                continue
            l = getattr(p, classname.lower() + 's', ())
            try:
                l.remove(obj)
            except ValueError:
                pass

    # Many-to-many relationships
    if isinstance(obj, neo.RecordingChannel):
        for rcg in obj.recordingchannelgroups:
            try:
                idx = rcg.recordingchannels.index(obj)
                if rcg.channel_indexes.shape[0] == len(rcg.recordingchannels):
                    rcg.channel_indexes = sp.delete(rcg.channel_indexes, idx)
                if rcg.channel_names.shape[0] == len(rcg.recordingchannels):
                    rcg.channel_names = sp.delete(rcg.channel_names, idx)
                rcg.recordingchannels.remove(obj)
            except ValueError:
                pass

    if isinstance(obj, neo.RecordingChannelGroup):
        for rc in obj.recordingchannels:
            try:
                rc.recordingchannelgroups.remove(obj)
            except ValueError:
                pass

    _handle_orphans(obj, remove_half_orphans)


def extract_spikes(train, signals, length, align_time):
    """ Extract spikes with waveforms from analog signals using a spike train. 
    Spikes that are too close to the beginning or end of the shortest signal
    to be fully extracted are ignored.

    :type train: :class:`neo.core.SpikeTrain`
    :param train: The spike times.
    :param sequence signals: A sequence of :class:`neo.core.AnalogSignal`
        objects from which the spikes are extracted. The waveforms of
        the returned spikes are extracted from these signals in the
        same order they are given.
    :type length: Quantity scalar
    :param length: The length of the waveform to extract as time scalar.
    :type align_time: Quantity scalar
    :param align_time: The alignment time of the spike times as time scalar.
        This is the time delta from the start of the extracted waveform
        to the exact time of the spike.
    :returns: A list of :class:`neo.core.Spike` objects, one for each
        time point in ``train``. All returned spikes include their
        ``waveform`` property.
    :rtype: list
    """
    if not signals:
        raise ValueError('No signals to extract spikes from')
    ref = signals[0]
    for s in signals[1:]:
        if ref.sampling_rate != s.sampling_rate:
            raise ValueError(
                'All signals for spike extraction need the same sampling rate')

    wave_unit = signals[0].units
    srate = signals[0].sampling_rate
    end = min(s.shape[0] for s in signals)

    aligned_train = train - align_time
    cut_samples = int((length * srate).simplified)

    st = sp.asarray((aligned_train * srate).simplified)

    # Find extraction epochs
    st_ok = (st >= 0) * (st < end - cut_samples)
    epochs = sp.vstack((st[st_ok], st[st_ok] + cut_samples)).T

    nspikes = epochs.shape[0]
    if not nspikes:
        return []

    # Create data
    data = sp.vstack([sp.asarray(s.rescale(wave_unit)) for s in signals])
    nc = len(signals)

    spikes = []
    for s in xrange(nspikes):
        waveform = sp.zeros((cut_samples, nc))
        for c in xrange(nc):
            waveform[:, c] = \
                data[c, epochs[s, 0]:epochs[s, 1]]
        spikes.append(neo.Spike(train[st_ok][s], waveform=waveform * wave_unit,
                                sampling_rate=srate))

    return spikes


import scipy as sp
import quantities as pq

from progress_indicator import ProgressIndicator
from . import SpykeException


def spike_amplitude_histogram(trains, num_bins, uniform_y_scale=True,
                              unit=pq.uV, progress=None):
    """ Return a spike amplitude histogram.

    The resulting is useful to assess the drift in spike amplitude over a
    longer recording. It shows histograms (one for each ``trains`` entry,
    e.g. segment) of maximum and minimum spike amplitudes.

    :param list trains: A list of lists of :class:`neo.core.SpikeTrain`
        objects. Each entry of the outer list will be one point on the
        x-axis (they could correspond to segments), all amplitude occurences
        of spikes contained in the inner list will be added up.
    :param int num_bins: Number of bins for the histograms.
    :param bool uniform_y_scale: If True, the histogram for each channel
        will use the same bins. Otherwise, the minimum bin range is computed
        separately for each channel.
    :param Quantity unit: Unit of Y-Axis.
    :param progress: Set this parameter to report progress.
    :type progress: :class:`.progress_indicator.ProgressIndicator`
    :return: A tuple with three values:

        * A three-dimensional histogram matrix, where the first dimension
          corresponds to bins, the second dimension to the entries of
          ``trains`` (e.g. segments) and the third dimension to channels.
        * A list of the minimum amplitude value for each channel (all values
          will be equal if ``uniform_y_scale`` is true).
        * A list of the maximum amplitude value for each channel (all values
          will be equal if ``uniform_y_scale`` is true).
    :rtype: (ndarray, list, list)
    """
    if not progress:
        progress = ProgressIndicator()

    num_channels = 1
    for t in trains:
        if not t:
            continue
        num_channels = t[0].waveforms.shape[2]
        break

    progress.set_ticks(2*len(trains))
    progress.set_status('Calculating Spike Amplitude Histogram')

    # Find maximum and minimum amplitudes on all channels
    up = [0] * num_channels
    down = [0] * num_channels
    for t in trains:
        for s in t:
            if s.waveforms is None:
                continue
            if s.waveforms.shape[2] != num_channels:
                raise SpykeException('All spikes need to have the same ' +
                                     'numer of channels for Spike Amplitude Histogram!')
            a = sp.asarray(s.waveforms.rescale(unit))
            u = a.max(1)
            d = a.min(1)
            for c in xrange(num_channels):
                up[c] = max(up[c], sp.stats.mstats.mquantiles(
                    u[:,c], [0.999])[0])
                down[c] = min(down[c], sp.stats.mstats.mquantiles(
                    d[:,c], [0.001])[0])
            progress.step()

    if uniform_y_scale:
        up = [max(up)] * num_channels
        down = [min(down)] * num_channels

    # Create histogram
    bins = [sp.linspace(down[c],up[c], num_bins+1)
            for c in xrange(num_channels)]
    hist = sp.zeros((num_bins, len(trains), num_channels))
    for i, t in enumerate(trains):
        for s in t:
            if s.waveforms is None:
                continue
            a = sp.asarray(s.waveforms.rescale(unit))
            upper = a.max(1)
            lower = a.min(1)
            for c in xrange(num_channels):
                hist[:,i,c] += sp.histogram(upper[:,c], bins[c])[0]
                hist[:,i,c] += sp.histogram(lower[:,c], bins[c])[0]
        progress.step()

    return hist, down, up



""" Contains a helper class for creating windows containing guiqwt plots.
"""

from PyQt4 import QtGui
from PyQt4.QtCore import Qt
from PyQt4.QtGui import (QDialog, QGridLayout, QToolBar, QHBoxLayout,
                         QVBoxLayout, QFrame, QWidget)
try:
    from PyQt4 import QtCore
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

from PyQt4.QtGui import QColor, QScrollArea, QWidget, QFrame
from guiqwt.baseplot import BasePlot
from guiqwt.curve import CurvePlot
from guiqwt.image import ImagePlot
from guiqwt.plot import PlotManager
from guiqwt.shapes import Marker
from guiqwt.curve import CurveItem
from guiqwt.tools import (SelectTool, RectZoomTool, BasePlotMenuTool,
                          DeleteItemTool, ItemListPanelTool,
                          AntiAliasingTool, AxisScaleTool, DisplayCoordsTool,
                          ExportItemDataTool, ItemCenterTool, SignalStatsTool,
                          ColormapTool, ReverseYAxisTool, AspectRatioTool,
                          ContrastPanelTool, XCSPanelTool, YCSPanelTool,
                          CrossSectionTool, AverageCrossSectionTool,
                          SaveAsTool, PrintTool, CopyToClipboardTool)
from guiqwt.signals import SIG_PLOT_AXIS_CHANGED
from guidata.configtools import get_icon

import icons_rc
import guiqwt_tools


# Monkeypatch curve and image plot so synchronizing axes works with all tools
def fixed_do_zoom_rect_view(self, *args, **kwargs):
    CurvePlot.old_do_zoom_rect_view(self, *args, **kwargs)
    self.emit(SIG_PLOT_AXIS_CHANGED, self)

CurvePlot.old_do_zoom_rect_view = CurvePlot.do_zoom_rect_view
CurvePlot.do_zoom_rect_view = fixed_do_zoom_rect_view


def fixed_do_autoscale(self, *args, **kwargs):
    CurvePlot.old_do_autoscale(self, *args, **kwargs)
    if not isinstance(self, ImagePlot):
        self.emit(SIG_PLOT_AXIS_CHANGED, self)

CurvePlot.old_do_autoscale = CurvePlot.do_autoscale
CurvePlot.do_autoscale = fixed_do_autoscale


def fixed_do_autoscale_image(self, *args, **kwargs):
    ImagePlot.old_do_autoscale(self, *args, **kwargs)
    self.emit(SIG_PLOT_AXIS_CHANGED, self)

ImagePlot.old_do_autoscale = ImagePlot.do_autoscale
ImagePlot.do_autoscale = fixed_do_autoscale_image


class PlotDialog(QDialog, PlotManager):
    """ Implements a dialog to which an arbitrary number of plots can be
    added.

    This class implements a `QDialog` with a number of plots on it. The
    axes of the plots can be arbitrarily synchronized and option checkboxes
    can be added which provide callbacks when the checkbox state changes.

    :param str wintitle: Title of the window.
    :param bool major_grid: Show major grid in plots.
    :param bool minor_grid: Show minor grid in plots.
    :param bool toolbar: Show toolbar.
    :param parent: Parent window.
    :param panels: A list of guiqwt panels to be added to the window.
    :param int min_plot_width: Default minimum width for plots.
    :param int min_plot_height: Default minimum height for plots.
    """

    def __init__(self, wintitle='Plot window', major_grid=True,
                 minor_grid=False, toolbar=True,  parent=None,
                 panels=None, min_plot_width=100, min_plot_height=75):
        QDialog.__init__(self, parent)
        self.setWindowFlags(Qt.Window)

        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(_fromUtf8(':/Application/Main')),
                       QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.setWindowIcon(icon)

        self.major_grid = major_grid
        self.minor_grid = minor_grid
        self.min_plot_width = min_plot_width
        self.min_plot_height = min_plot_height

        # WidgetMixin copy
        PlotManager.__init__(self, main=self)
        
        self.main_layout = QVBoxLayout(self)
        self.color_layout = QHBoxLayout()
        self.plot_layout = QGridLayout()
        self.plot_layout.setMargin(0)
        self.plot_scroll_widget = QWidget()
        self.plot_scroll_area = QScrollArea()
        self.plot_scroll_area.setFrameShape(QFrame.NoFrame)
        self.plot_scroll_area.setWidgetResizable(True)
        self.option_layout = QHBoxLayout()

        self.plot_widget = None

        if panels is not None:
            for panel in panels:
                self.add_panel(panel)

        self.toolbar = QToolBar('Tools')
        if not toolbar:
            self.toolbar.hide()

        # Configuring widget layout
        self._setup_widget_properties(wintitle=wintitle, icon=icon)
        self._setup_widget_layout()
        
        # Options
        self.option_callbacks = {}
        self.legend = None
        self.axis_syncplots = {}
        
    def _setup_widget_properties(self, wintitle, icon):
        self.setWindowTitle(wintitle)
        if isinstance(icon, basestring):
            icon = get_icon(icon)
        self.setWindowIcon(icon)
        self.setMinimumSize(320, 240)
        self.resize(720, 540)
        
    def _setup_widget_layout(self):
        self.main_layout.addWidget(self.toolbar)
        self.main_layout.addLayout(self.color_layout)
        self.main_layout.addWidget(self.plot_scroll_area)
        self.plot_scroll_area.setWidget(self.plot_scroll_widget)
        self.plot_scroll_widget.setLayout(self.plot_layout)
        self.main_layout.addLayout(self.option_layout)
        self.setLayout(self.main_layout)
        
    def add_custom_curve_tools(self, antialiasing=True,
                               activate_zoom=True,
                               signal_stats=False):
        """ Adds typically needed curve tools to the window.

        :param bool antialiasing: Determines if the antialiasing tool is
            added.
        :param bool activate_zoom: Determines if the zoom tool is activated
            initially (otherwise, the selection tool will be activated).
        :param bool signal_stats: Determines if the signal stats tool is
            available.
        """
        self.add_toolbar(self.toolbar)

        t = self.add_tool(SelectTool)
        if not activate_zoom:
            self.set_default_tool(t)
        self.add_tool(BasePlotMenuTool, "item")
        self.add_tool(ExportItemDataTool)
        try:  # Old versions of guiqwt and spyderlib do not support this
            import spyderlib.widgets.objecteditor
            from guiqwt.tools import EditItemDataTool
            self.add_tool(EditItemDataTool)
        except ImportError:
            pass
        self.add_tool(ItemCenterTool)
        self.add_tool(DeleteItemTool)

        self.add_separator_tool()
        t = self.add_tool(RectZoomTool)
        if activate_zoom:
            self.set_default_tool(t)
        self.add_tool(guiqwt_tools.HomeTool)
        self.add_tool(guiqwt_tools.PanTool)

        self.add_separator_tool()
        self.add_tool(BasePlotMenuTool, "grid")
        self.add_tool(BasePlotMenuTool, "axes")
        self.add_tool(DisplayCoordsTool)
        if self.get_itemlist_panel():
            self.add_tool(ItemListPanelTool)

        if signal_stats:
            self.add_separator_tool()
            self.add_tool(SignalStatsTool)

        if antialiasing:
            self.add_tool(AntiAliasingTool)
        self.add_tool(AxisScaleTool)
        self.add_separator_tool()
        self.add_tool(SaveAsTool)
        self.add_tool(CopyToClipboardTool)
        self.add_tool(PrintTool)
        self.add_tool(guiqwt_tools.HelpTool)
        self.add_separator_tool()
        self.get_default_tool().activate()

    def add_custom_image_tools(self, activate_zoom=True):
        """ Adds typically needed image tools to the window.
        """
        self.add_toolbar(self.toolbar)

        t = self.add_tool(SelectTool)
        if not activate_zoom:
            self.set_default_tool(t)
        self.add_tool(BasePlotMenuTool, "item")
        self.add_tool(ExportItemDataTool)
        try:  # Old versions of guiqwt and spyderlib do not support this
            import spyderlib.widgets.objecteditor
            from guiqwt.tools import EditItemDataTool
            self.add_tool(EditItemDataTool)
        except ImportError:
            pass
        self.add_tool(ItemCenterTool)
        self.add_tool(DeleteItemTool)

        self.add_separator_tool()
        t = self.add_tool(RectZoomTool)
        if activate_zoom:
            self.set_default_tool(t)
        self.add_tool(guiqwt_tools.HomeTool)
        self.add_tool(guiqwt_tools.PanTool)

        self.add_separator_tool()
        self.add_tool(BasePlotMenuTool, "grid")
        self.add_tool(BasePlotMenuTool, "axes")
        self.add_tool(DisplayCoordsTool)
        if self.get_itemlist_panel():
            self.add_tool(ItemListPanelTool)

        self.add_separator_tool()
        self.add_tool(ColormapTool)
        self.add_tool(ReverseYAxisTool)
        self.add_tool(AspectRatioTool)
        if self.get_contrast_panel():
            self.add_tool(ContrastPanelTool)
        if self.get_xcs_panel() and self.get_ycs_panel():
            self.add_tool(XCSPanelTool)
            self.add_tool(YCSPanelTool)
            self.add_tool(CrossSectionTool)
            self.add_tool(AverageCrossSectionTool)

        self.add_separator_tool()
        self.add_tool(SaveAsTool)
        self.add_tool(CopyToClipboardTool)
        self.add_tool(PrintTool)
        self.add_tool(guiqwt_tools.HelpTool)
        self.add_separator_tool()
        self.get_default_tool().activate()
    
    def add_option(self, name, change_callback, active=False):
        """ Add an option (using a checkbox) to the window.

        :param str name: The name of the option.
        :param func change_callback: A function accepting the new state as
            a parameter. The function will be called whenever the state
            of the option changes.
        :param bool active: Determines if the option is activated initially.
        """
        checkBox = QtGui.QCheckBox(self)
        checkBox.setChecked(active)
        checkBox.setText(name)
        checkBox.stateChanged.connect(self._option_callback)
        self.option_callbacks[checkBox] = change_callback
        self.option_layout.addWidget(checkBox)

    def add_x_synchronization_option(self, active, ids=None):
        """ Offer an option for X axes synchronization. This method should
        be called after show(), so that a proper initial synchronization
        can be performed.

        :param bool active: Determines whether the axes are synchronized
            initially.
        :param sequence ids: List of plot ids to synchronize.
        """
        self.axis_syncplots[BasePlot.X_BOTTOM] = ids
        if active and ids:
            self.synchronize_axis(BasePlot.X_BOTTOM)
        self.add_option('Synchronize X Axes',
                        PlotDialog._synchronization_option_x, active)

    def add_y_synchronization_option(self, active, ids=None):
        """ Offer an option for Y axes synchronization. This method should
        be called after show(), so that a proper initial synchronization
        can be performed.

        :param bool active: Determines whether the axes are synchronized
            initially
        :param sequence ids: List of plot ids to synchronize.
        """
        self.axis_syncplots[BasePlot.Y_LEFT] = ids
        if active and ids:
            self.synchronize_axis(BasePlot.Y_LEFT)
        self.add_option('Synchronize Y Axes',
                        PlotDialog._synchronization_option_y, active)

    def _synchronization_option_x(self, state):
        """ Callback for x-axis synchronization
        """
        if state:
            self.synchronize_axis(BasePlot.X_BOTTOM)
        else:
            self.unsynchronize_axis(BasePlot.X_BOTTOM)

    def _synchronization_option_y(self, state):
        """ Callback for y-axis synchronization
        """
        if state:
            self.synchronize_axis(BasePlot.Y_LEFT)
        else:
            self.unsynchronize_axis(BasePlot.Y_LEFT)

    def replace_colors(self, replace_list):
        """ Replace colors of items in all plots.

            This can be useful when changing the background color to black
            and black items should be drawn in white:
            ``replace_colors([('#000000', '#ffffff']))``

        :param list replace_list: A list of tuples of Qt color names. The
            first color in each tuple is replaced by the second color.
        """
        for plot in self.plots.itervalues():
            for i in plot.get_items():
                if isinstance(i, CurveItem):
                    pen = i.pen()
                elif isinstance(i, Marker):
                    pen = i.linePen()
                else:
                    continue
                for color in replace_list:
                    c1 = QColor(color[0])
                    c2 = QColor(color[1])
                    if pen.color() != c1:
                        continue
                    pen.setColor(c2)
                    break
                if isinstance(i, CurveItem):
                    i.setPen(pen)
                elif isinstance(i, Marker):
                    i.setLinePen(pen)
            plot.replot()

    def set_background_color(self, color):
        """ Set the background color for all plots.

        :param str color: A Qt color name (e.g. '#ff0000')
        """
        for p in self.plots.itervalues():
            p.setCanvasBackground(QColor(color))
            p.replot()
        
    def add_unit_color(self, color, name='Unit color:'):
        """ Create a small legend on top of the window with only one entry.

        :param str color: A Qt color name (e.g. '#ff0000')
        :param str name: The name of the legend item. It will be displayed
            on the left of the color.
        """
        label = QtGui.QLabel(self)
        label.setText(name)
        label.setAlignment(Qt.AlignRight)
        self.color_layout.addWidget(label)
        label = QtGui.QLabel(self)
        label.setStyleSheet('background-color:' + str(color))
        label.setFrameShape(QFrame.StyledPanel)
        label.setMaximumWidth(80)
        self.color_layout.addWidget(label)

    def add_custom_label(self, legend_string):
        """ Add a label on the right of the plots

        :param str legend_string: An arbitrary string (which can contain
            newlines) to display on the right of the plots
        """
        label = QtGui.QLabel(self)
        label.setText(legend_string)
        self.plot_layout.addWidget(
            label, 0, self.plot_layout.columnCount(), -1, 1)

    def add_color_legend(self, legend, show_option=None):
        """ Create a legend on the right of the plots with colors and names.

        :param sequence legend: List of (color, text) tuples, where `color`
            is a Qt color name (e.g. '#ff0000') and `text` is the
            corresponding text to display in the legend.
        :param bool show_option: Determines whether a toggle for the legend
            will be shown (if the parameter is not ``None``) and if the legend
            is visible initially (``True``/``False``).
        """
        widget = QWidget(self)
        layout = QGridLayout(widget)
        widget.setLayout(layout)
        for l in legend:
            label = QtGui.QLabel(self)
            label.setStyleSheet('background-color:' + str(l[0]))
            label.setFrameShape(QFrame.StyledPanel)
            label.setMaximumWidth(80)
            label.setMaximumHeight(12)
            layout.addWidget(label, layout.rowCount(), 0, 1, 1)
            label = QtGui.QLabel(self)
            label.setText(l[1])
            layout.addWidget(label, layout.rowCount() - 1, 1, 1, 1)

        self.plot_layout.addWidget(
            widget, 0, self.plot_layout.columnCount(), -1, 1)
        if show_option is not None:
            widget.setVisible(show_option)
            self.add_option(
                'Show Legend Sidebar', lambda w, s: widget.setVisible(s),
                show_option)
    
    def add_legend_option(self, legends, active):
        """ Create a user option to show or hide a list of legend objects.

        :param sequence legends: The legend objects affected by the option.
        :param bool active: Determines whether the legends will be visible
            initially.
        """
        self.legends = legends
        self._set_legend_visibility(active)
        self.add_option('Show legend', self._legend_callback, active)
        if active:
            self._set_legend_visibility(True)
        
    def _option_callback(self, state):
        self.option_callbacks[self.sender()](self, state)

    #noinspection PyUnusedLocal
    def _legend_callback(self, win, state):
        self._set_legend_visibility(state > 0)
    
    def _set_legend_visibility(self, visible):
        for p in self.plots.itervalues():
            for l in self.legends:
                p.set_item_visible(l, visible)

    def add_plot_widget(self, plot_widget, plot_id, row=-1, column=0,
                        min_plot_width=None, min_plot_height=None):
        """ Adds a guiqwt plot to the window.

        :param plot_widget: The plot to add.
        :type plot_widget: guiqwt plot widget
        :param int plot_id: The id of the new plot.
        :param int row: The row of the new plot. If this is -1, the new plot
            will be added in a new row (if `column` is 0) or
            in the last row.
        :param int column: The column of the new plot.
        :param int min_plot_width: The minimum width of this plot. If
            ``None``, the default minimum width for this dialog
            is used.
        :param int max_plot_height: The minimum height of this plot. If
            ``None``, the default minimum height for this dialog
            is used.
        """
        if row == -1:
            if column == 0:
                row = self.plot_layout.rowCount()
            else:
                row = self.plot_layout.rowCount() - 1
        pw = min_plot_width
        if pw is None:
            pw = self.min_plot_width
        ph = min_plot_height
        if ph is None:
            ph = self.min_plot_height
        plot_widget.setMinimumSize(pw, ph)
        self.plot_layout.addWidget(plot_widget, row, column)
        new_plot = plot_widget.plot
        self.add_plot(new_plot, plot_id)
        
    def synchronize_axis(self, axis, plots=None):
        if plots is None:
            if axis in self.axis_syncplots:
                plots = self.axis_syncplots[axis]
            else:
                plots = self.plots.keys()
        if len(plots) < 1:
            return

        PlotManager.synchronize_axis(self, axis, plots)

        # Find interval that needs to be shown in order to include all
        # currently shown parts in the synchronized plots
        plot_objects = [self.plots[p] for p in plots]
        lb = min((p.axisScaleDiv(axis).lowerBound() for p in plot_objects))
        ub = max((p.axisScaleDiv(axis).upperBound() for p in plot_objects))
        for p in plot_objects:
            p.setAxisScale(axis, lb, ub)
            p.replot()

    def unsynchronize_axis(self, axis, plots=None):
        if plots is None:
            if axis in self.axis_syncplots:
                plots = self.axis_syncplots[axis]
            else:
                plots = self.plots.keys()
        for plot_id in plots:
            if not plot_id in self.synchronized_plots:
                continue
            synclist = self.synchronized_plots[plot_id]
            for plot2_id in plots:
                if plot_id == plot2_id:
                    continue
                item = (axis, plot2_id)
                if item in synclist:
                    synclist.remove(item)
        
    def plot_axis_changed(self, plot):
        ids = [k for k, p in self.plots.iteritems() if p == plot]
        if len(ids) < 1:
            return
        plot_id = ids[0]
        if plot_id not in self.synchronized_plots:
            return
        for (axis, other_plot_id) in self.synchronized_plots[plot_id]:
            scalediv = plot.axisScaleDiv(axis)
            other = self.get_plot(other_plot_id)
            lb = scalediv.lowerBound()
            ub = scalediv.upperBound()
            other.setAxisScale(axis, lb, ub)
            other.replot()

    def set_plot_title(self, plot, title):
        """ Set the title of a guiqwt plot and use the same font as for the
            rest of the window.

        :param plot: The plot for which the title is set.
        :param str title: The new title of the plot.
        """
        plot.setTitle(title)
        l = plot.titleLabel()
        l.setFont(self.font())
        plot.setTitle(l.text())

    def show(self):
        for p in self.plots.itervalues():
            if not self.minor_grid:
                p.grid.gridparam.min_xenabled = False
                p.grid.gridparam.min_yenabled = False
            if not self.major_grid:
                p.grid.gridparam.maj_xenabled = False
                p.grid.gridparam.maj_yenabled = False
            p.grid.update_params()
        super(PlotDialog, self).show()



import neo
import quantities as pq
import scipy as sp


def create_empty_spike_train(t_start=0.0 * pq.s, t_stop=10.0 * pq.s):
    return neo.SpikeTrain(sp.array([]) * pq.s, t_start=t_start, t_stop=t_stop)

def arange_spikes(t_start, t_stop=None, t_step=1.0 * pq.s):
    """ Arranges equally spaced spikes in a spike train over an interval. There
    will be no spikes at the interval boundaries.

    :param t_start: The time point of the interval if `t_stop` is not `None`.
        Otherwise, it will be end point of the interval and the start point will
        be set to 0s.
    :type t_start: Quantity scalar
    :param t_stop: The end point of the interval.
    :type t_stop: Quantity scalar
    :param t_step: Spacing between the spikes.
    :type t_step: Quantity scalar
    :returns The arranged spike train.
    :rtype: :class:`neo.SpikeTrain`
    """

    if t_stop is None:
        t_stop = t_start
        t_start = 0.0 * pq.s

    t_start.units = t_step.units
    t_stop.units = t_step.units

    spike_times = sp.arange(t_start + t_step, t_stop, t_step) * t_step.units
    return neo.SpikeTrain(spike_times, t_start=t_start, t_stop=t_stop)



import os
import sys
from copy import copy
from collections import OrderedDict
import traceback
import atexit

import neo

from data_provider import DataProvider
from .. import conversions as convert


class NeoDataProvider(DataProvider):
    """ Base class for data providers using NEO"""

    # Dictionary of block lists, indexed by (filename, block index) tuples
    loaded_blocks = {}
    # Dictionary of index in file, indexed by block object
    block_indices = {}
    # Dictionary of io, indexed by block object
    block_ios = {}
    # Dictionary of io (IO name, read paramters) tuples for loaded blocks
    block_read_params = {}
    # Mode for data lazy loading:
    # 0 - Full load
    # 1 - Lazy load
    # 2 - Caching lazy load
    data_lazy_mode = 0
    # Mode for lazy cascade
    cascade_lazy = False
    # Forced IO class for all files. If None, determine by file extension.
    forced_io = None
    # Active IO read parameters (dictionary indexed by IO class)
    io_params = {}

    def __init__(self, name, progress):
        super(NeoDataProvider, self).__init__(name, progress)

    @classmethod
    def clear(cls):
        """ Clears cached blocks
        """
        cls.loaded_blocks.clear()
        cls.block_indices.clear()
        cls.block_read_params.clear()

        ios = set()
        for io in cls.block_ios.itervalues():
            if io in ios:
                continue
            if hasattr(io, 'close'):
                io.close()
                ios.add(io)
        cls.block_ios.clear()

    @classmethod
    def get_block(cls, filename, index, lazy=None, force_io=None,
                  read_params=None):
        """ Return the block at the given index in the specified file.

        :param str filename: Path to the file from which to load the block.
        :param int index: The index of the block in the file.
        :param int lazy: Override global lazy setting if not ``None``:
            0 regular load, 1 lazy load, 2 caching lazy load.
        :param force_io: Override global forced_io for the Neo IO class
            to use when loading the file. If ``None``, the global
            forced_io is used.
        :param dict read_params: Override read parameters for the IO that
            will load the block. If ``None``, the global io_params are
            used.
        """
        if lazy is None:
            lazy = cls.data_lazy_mode > 0
        else:
            lazy = lazy > 0
        if force_io is None:
            force_io = cls.forced_io

        if filename in cls.loaded_blocks:
            return cls.loaded_blocks[filename][index]
        io, blocks = cls._load_neo_file(filename, lazy, force_io, read_params)
        if io and not lazy and not cls.cascade_lazy and hasattr(io, 'close'):
            io.close()
        if blocks is None:
            return None
        return blocks[index]

    @classmethod
    def get_blocks(cls, filename, lazy=None, force_io=None,
                   read_params=None):
        """ Return a list of blocks loaded from the specified file

        :param str filename: Path to the file from which to load the blocks.
        :param int lazy: Override global lazy setting if not ``None``:
            0 regular load, 1 lazy load, 2 caching lazy load.
        :param force_io: Override global forced_io for the Neo IO class
            to use when loading the file. If ``None``, the global
            forced_io is used.
        :param dict read_params: Override read parameters for the IO that
            will load the block. If ``None``, the global io_params are
            used.
        """
        if lazy is None:
            lazy = cls.data_lazy_mode > 0
        else:
            lazy = lazy > 0
        if force_io is None:
            force_io = cls.forced_io

        if filename in cls.loaded_blocks:
            return cls.loaded_blocks[filename]
        io, blocks = cls._load_neo_file(filename, lazy, force_io, read_params)
        if io and not lazy and not cls.cascade_lazy and hasattr(io, 'close'):
            io.close()
        return blocks

    @classmethod
    def _load_neo_file(cls, filename, lazy, force_io, read_params):
        """ Returns a NEO io object and a list of contained blocks for a
        file name. This function also caches all loaded blocks

        :param str filename: The full path of the file (relative or absolute).
        :param bool lazy: Determines if lazy mode is used for Neo io.
        :param force_io: IO class to use for loading. If None, determined
            by file extension or through trial and error for directories.
        :param dict read_params: Override read parameters for the IO that
            will load the block. If ``None``, the global io_params are
            used.
        """
        cascade = 'lazy' if cls.cascade_lazy else True
        if os.path.isdir(filename):
            if force_io:
                try:
                    n_io = force_io(filename)
                    if read_params is None:
                        rp = cls.io_params.get(force_io, {})
                    else:
                        rp = read_params
                    content = n_io.read(lazy=lazy, cascade=cascade, **rp)
                    if force_io == neo.TdtIO and \
                            isinstance(content, neo.Block) and \
                            not content.segments:
                        # TdtIO can produce empty blocks for invalid dirs
                        sys.stderr.write(
                            'Could not load any blocks from "%s"' % filename)
                        return None, None

                    return cls._content_loaded(
                        content, filename, lazy, n_io, rp)
                except Exception, e:
                    sys.stderr.write(
                        'Load error for directory "%s":\n' % filename)
                    tb = sys.exc_info()[2]
                    while not ('self' in tb.tb_frame.f_locals and
                               tb.tb_frame.f_locals['self'] == n_io):
                        if tb.tb_next is not None:
                            tb = tb.tb_next
                        else:
                            break
                    traceback.print_exception(type(e), e, tb)
            else:
                for io in neo.io.iolist:
                    if io.mode == 'dir':
                        try:
                            n_io = io(filename)
                            if read_params is None:
                                rp = cls.io_params.get(force_io, {})
                            else:
                                rp = read_params
                            content = n_io.read(lazy=lazy, cascade=cascade, **rp)
                            if io == neo.TdtIO and \
                                    isinstance(content, neo.Block) and \
                                    not content.segments:
                                # TdtIO can produce empty blocks for invalid dirs
                                continue

                            return cls._content_loaded(
                                content, filename, lazy, n_io, rp)
                        except Exception, e:
                            sys.stderr.write(
                                'Load error for directory "%s":\n' % filename)
                            tb = sys.exc_info()[2]
                            while not ('self' in tb.tb_frame.f_locals and
                                       tb.tb_frame.f_locals['self'] == n_io):
                                if tb.tb_next is not None:
                                    tb = tb.tb_next
                                else:
                                    break
                            traceback.print_exception(type(e), e, tb)
        else:
            if force_io:
                if read_params is None:
                    rp = cls.io_params.get(force_io, {})
                else:
                    rp = read_params
                return cls._load_file_with_io(filename, force_io, lazy, rp)

            extension = filename.split('.')[-1]
            for io in neo.io.iolist:
                if extension in io.extensions:
                    if read_params is None:
                        rp = cls.io_params.get(io, {})
                    else:
                        rp = read_params
                    return cls._load_file_with_io(filename, io, lazy, rp)

        return None, None

    @classmethod
    def _content_loaded(cls, content, filename, lazy, n_io, read_params):
        if isinstance(content, neo.Block):  # Neo 0.2.1
            cls.block_indices[content] = 0
            cls.loaded_blocks[filename] = [content]
            cls.block_read_params[content] = (type(n_io).__name__, read_params)
            if lazy or cls.cascade_lazy:
                cls.block_ios[content] = n_io
            return n_io, [content]

        # Neo >= 0.3.0, read() returns a list of blocks
        blocks = content
        for i, b in enumerate(blocks):
            cls.block_indices[b] = i
            cls.block_read_params[b] = (type(n_io).__name__, read_params)
            if lazy or cls.cascade_lazy:
                cls.block_ios[b] = n_io

        cls.loaded_blocks[filename] = blocks
        return n_io, blocks

    @classmethod
    def _load_file_with_io(cls, filename, io, lazy, read_params):
        if io == neo.NeoHdf5IO:
            # Fix unicode problem with pyinstaller
            if hasattr(sys, 'frozen'):
                filename = filename.encode('UTF-8')

        n_io = io(filename=filename)

        if read_params is None:
            rp = cls.io_params.get(io, {})
        else:
            rp = read_params

        try:
            cascade = 'lazy' if cls.cascade_lazy else True
            if hasattr(io, 'read_all_blocks'):  # Neo 0.2.1
                content = n_io.read_all_blocks(lazy=lazy, cascade=cascade, **rp)
            else:
                content = n_io.read(lazy=lazy, cascade=cascade, **rp)

            return cls._content_loaded(content, filename, lazy, n_io, rp)
        except Exception, e:
            sys.stderr.write(
                'Load error for file "%s":\n' % filename)
            tb = sys.exc_info()[2]
            while not ('self' in tb.tb_frame.f_locals and
                               tb.tb_frame.f_locals['self'] == n_io):
                if tb.tb_next is not None:
                    tb = tb.tb_next
                else:
                    break
            traceback.print_exception(type(e), e, tb)

        return None, None

    @classmethod
    def _get_data_from_viewer(cls, viewer):
        """ Return a dictionary with selection information from viewer
        """
        # The links in this data format are based list indices
        data = {}
        data['type'] = 'Neo'

        # Block entry: (Index of block in file, file location of block,
        # block IO class name, block IO read parameters)
        block_list = []
        block_indices = {}
        selected_blocks = viewer.neo_blocks()
        block_files = viewer.neo_block_file_names()
        for b in selected_blocks:
            block_indices[b] = len(block_list)
            block_list.append([NeoDataProvider.block_indices[b],
                               block_files[b],
                               cls.block_read_params[b][0],
                               cls.block_read_params[b][1]])
        data['blocks'] = block_list

        # Recording channel group entry:
        # (Index of rcg in block, index of block)
        rcg_list = []
        rcg_indices = {}
        selected_rcg = viewer.neo_channel_groups()
        for rcg in selected_rcg:
            rcg_indices[rcg] = len(rcg_list)
            idx = rcg.block.recordingchannelgroups.index(rcg)
            rcg_list.append([idx, block_indices[rcg.block]])
        data['channel_groups'] = rcg_list

        # Recording channel entry: (Index of channel in rcg, index of rcg)
        # There can be multiple channel entries for one channel object, if
        # it is part of multiple channel groups
        channel_list = []
        selected_channels = viewer.neo_channels()
        for c in selected_channels:
            for rcg in c.recordingchannelgroups:
                if rcg in rcg_indices:
                    idx = rcg.recordingchannels.index(c)
                    channel_list.append([idx, rcg_indices[rcg]])
        data['channels'] = channel_list

        # Segment entry: (Index of segment in block, index of block)
        segment_list = []
        segment_indices = {}
        selected_segments = viewer.neo_segments()
        for s in selected_segments:
            segment_indices[s] = len(segment_list)
            idx = s.block.segments.index(s)
            segment_list.append([idx, block_indices[s.block]])
        data['segments'] = segment_list

        # Unit entry: (Index of unit in rcg, index of rcg)
        unit_list = []
        selected_units = viewer.neo_units()
        for u in selected_units:
            segment_indices[u] = len(segment_list)
            rcg_id = None if u.recordingchannelgroup is None \
                else u.recordingchannelgroup.units.index(u)
            rcg = rcg_indices[u.recordingchannelgroup] \
                if u.recordingchannelgroup else None
            unit_list.append([rcg_id, rcg])
        data['units'] = unit_list

        return data

    @staticmethod
    def find_io_class(name):
        """ Return the Neo IO class with a given name.

        :param str name: Class name of the desired IO class.
        """
        for io in neo.io.iolist:
            if io.__name__ == name:
                return io
        return None

    def _active_block(self, old):
        """ Return a copy of all selected elements in the given block.
        Only container objects are copied, data objects are linked.

        Needs to load all lazily loaded objects and will cache them
        regardless of current lazy_mode,
        """
        block = copy(old)

        block.segments = []
        selected_segments = set(self.segments() + [None])
        selected_rcgs = set(self.recording_channel_groups() + [None])
        selected_channels = set(self.recording_channels() + [None])
        selected_units = set(self.units() + [None])

        for s in old.segments:
            if s in selected_segments:
                segment = copy(s)
                segment.analogsignals = [self._load_lazy_object(sig, True)
                                         for sig in s.analogsignals
                                         if sig.recordingchannel
                                         in selected_channels]
                segment.analogsignalarrays = [
                    self._load_lazy_object(asa, True)
                    for asa in s.analogsignalarrays
                    if asa.recordingchannelgroup in selected_rcgs]
                segment.irregularlysampledsignals = [
                    self._load_lazy_object(iss, True)
                    for iss in s.irregularlysampledsignals
                    if iss.recordingchannel in selected_channels]
                segment.spikes = [self._load_lazy_object(sp, True)
                                  for sp in s.spikes
                                  if sp.unit in selected_units]
                segment.spiketrains = [self._load_lazy_object(st, True)
                                       for st in s.spiketrains
                                       if st.unit in selected_units]
                segment.block = block
                block.segments.append(segment)

        block.recordingchannelgroups = []
        for old_rcg in old.recordingchannelgroups:
            if old_rcg in selected_rcgs:
                rcg = copy(old_rcg)
                rcg.analogsignalarrays = [
                    self._load_lazy_object(asa, True)
                    for asa in old_rcg.analogsignalarrays
                    if asa.segment in selected_segments]

                rcg.recordingchannels = []
                for c in old_rcg.recordingchannels:
                    if not c in selected_channels:
                        continue
                    channel = copy(c)
                    channel.analogsignals = [
                        self._load_lazy_object(sig, True)
                        for sig in c.analogsignals
                        if sig.segment in selected_segments]
                    channel.irregularlysampledsignals = [
                        self._load_lazy_object(iss, True)
                        for iss in c.irregularlysampledsignals
                        if iss.segment in selected_segments]
                    channel.recordingchannelgroups = copy(
                        c.recordingchannelgroups)
                    channel.recordingchannelgroups.insert(
                        channel.recordingchannelgroups.index(old_rcg), rcg)
                    channel.recordingchannelgroups.remove(old_rcg)
                    rcg.recordingchannels.append(channel)

                rcg.units = []
                for u in old_rcg.units:
                    if not u in selected_units:
                        continue

                    unit = copy(u)
                    unit.spikes = [self._load_lazy_object(sp, True)
                                   for sp in u.spikes
                                   if sp.segment in selected_segments]
                    unit.spiketrains = [self._load_lazy_object(st, True)
                                        for st in u.spiketrains
                                        if st.segment in selected_segments]
                    unit.recordingchannelgroup = rcg
                    rcg.units.append(unit)

                rcg.block = block
                block.recordingchannelgroups.append(rcg)

        return block

    def _get_object_io(self, o):
        """ Find the IO for an object. Return ``None`` if no IO exists.
        """
        if o.segment:
            return self.block_ios.get(o.segment.block, None)
        if hasattr(object, 'recordingchannelgroups'):
            if o.recordingchannelgroups:
                return self.block_ios.get(
                    o.recordingchannelgroups[0].block, None)
        if hasattr(object, 'recordingchannel'):
            c = o.recordingchannel
            if c.recordingchannelgroups:
                return self.block_ios.get(
                    c.recordingchannelgroups[0].block, None)
        return None

    def _load_lazy_object(self, o, change_links=False):
        """ Return a loaded version of a lazily loaded object. The IO
        needs a ``read_lazy_object`` that takes a lazily loaded data object
        as parameter method for this to work.

        :param o: The object to load.
        :param bool change_links: If ``True``, replace the old object
            in the hierarchy.
        """
        if not hasattr(o, 'lazy_shape'):
            return o

        io = self._get_object_io(o)

        if io:
            if hasattr(io, 'load_lazy_object'):
                ret = io.load_lazy_object(o)
            elif isinstance(io, neo.io.NeoHdf5IO):
                ret = io.get(o.hdf5_path, cascade=False, lazy=False)
            else:
                return o

            ret.segment = o.segment
            if hasattr(o, 'recordingchannelgroup'):
                ret.recordingchannelgroup = o.recordingchannelgroup
            elif hasattr(o, 'recordingchannel'):
                ret.recordingchannel = o.recordingchannel
            elif hasattr(o, 'unit'):
                ret.unit = o.unit

            if change_links:
                name = type(o).__name__.lower() + 's'
                l = getattr(o.segment, name)
                try:
                    l[l.index(o)] = ret
                except ValueError:
                    l.append(ret)

                l = None
                if hasattr(o, 'recordingchannelgroup'):
                    l = getattr(o.recordingchannelgroup, name)
                elif hasattr(o, 'recordingchannel'):
                    l = getattr(o.recordingchannel, name)
                elif hasattr(o, 'unit'):
                    l = getattr(o.unit, name)
                if l is not None:
                    try:
                        l[l.index(o)] = ret
                    except ValueError:
                        l.append(ret)

            return ret
        return o

    def _load_object_list(self, objects):
        """ Return a list of loaded objects for a list of (potentially)
        lazily loaded objects.
        """
        ret = []
        for o in objects:
            ret.append(self._load_lazy_object(o, self.data_lazy_mode > 1))
        return ret

    def _load_object_dict(self, objects):
        """ Return a dictionary (without changing indices) of loaded
        objects for a dictionary of (potentially) lazily loaded objects.
        """
        for k, v in objects.items():
            if isinstance(v, list):
                objects[k] = self._load_object_list(v)
            elif isinstance(v, dict):
                for ik, iv in v.items():
                    v[ik] = self._load_lazy_object(iv, self.data_lazy_mode > 1)
            else:
                raise ValueError(
                    'Only dicts or lists are supported as dictionary values!')
        return objects

    def selection_blocks(self):
        """ Return a list of selected blocks.
        """
        return [self._active_block(b) for b in self.blocks()]

    def spike_trains(self):
        """ Return a list of :class:`neo.core.SpikeTrain` objects.
        """
        trains = []
        units = set(self.units())
        for s in self.segments():
            trains.extend([t for t in s.spiketrains if t.unit in units or
                           t.unit is None])
        for u in self.units():
            trains.extend([t for t in u.spiketrains if t.segment is None])

        return self._load_object_list(trains)

    def spike_trains_by_unit(self):
        """ Return a dictionary (indexed by Unit) of lists of
        :class:`neo.core.SpikeTrain` objects.
        """
        trains = OrderedDict()
        segments = set(self.segments())
        for u in self.units():
            st = [t for t in u.spiketrains if t.segment in segments or
                  t.segment is None]
            if st:
                trains[u] = st

        nonetrains = []
        for s in self.segments():
            nonetrains.extend([t for t in s.spiketrains if t.unit is None])
        if nonetrains:
            trains[self.no_unit] = nonetrains

        return self._load_object_dict(trains)

    def spike_trains_by_segment(self):
        """ Return a dictionary (indexed by Segment) of lists of
        :class:`neo.core.SpikeTrain` objects.
        """
        trains = OrderedDict()
        units = self.units()
        for s in self.segments():
            st = [t for t in s.spiketrains if t.unit in units or
                  t.unit is None]
            if st:
                trains[s] = st

        nonetrains = []
        for u in self.units():
            nonetrains.extend([t for t in u.spiketrains if t.segment is None])
        if nonetrains:
            trains[self.no_segment] = nonetrains

        return self._load_object_dict(trains)

    def spike_trains_by_unit_and_segment(self):
        """ Return a dictionary (indexed by Unit) of dictionaries
        (indexed by Segment) of :class:`neo.core.SpikeTrain` objects.
        """
        trains = OrderedDict()
        segments = self.segments()
        for u in self.units():
            for s in segments:
                segtrains = [t for t in u.spiketrains if t.segment == s]
                if segtrains:
                    if u not in trains:
                        trains[u] = OrderedDict()
                    trains[u][s] = segtrains[0]
            nonetrains = [t for t in u.spiketrains if t.segment is None]
            if nonetrains:
                if u not in trains:
                    trains[u] = OrderedDict()
                trains[u][self.no_segment] = nonetrains[0]

        nonetrains = OrderedDict()
        for s in self.segments():
            segtrains = [t for t in s.spiketrains if t.unit is None]
            if segtrains:
                nonetrains[s] = segtrains[0]
        if nonetrains:
            trains[self.no_unit] = nonetrains

        return self._load_object_dict(trains)

    def spikes(self):
        """ Return a list of :class:`neo.core.Spike` objects.
        """
        spikes = []
        units = self.units()
        for s in self.segments():
            spikes.extend([t for t in s.spikes if t.unit in units or
                           t.unit is None])
        for u in self.units():
            spikes.extend([t for t in u.spikes if t.segment is None])

        return self._load_object_list(spikes)

    def spikes_by_unit(self):
        """ Return a dictionary (indexed by Unit) of lists of
        :class:`neo.core.Spike` objects.
        """
        spikes = OrderedDict()
        segments = self.segments()
        for u in self.units():
            sp = [t for t in u.spikes if t.segment in segments or
                  t.segment is None]
            if sp:
                spikes[u] = sp

        nonespikes = []
        for s in self.segments():
            nonespikes.extend([t for t in s.spikes if t.unit is None])
        if nonespikes:
            spikes[self.no_unit] = nonespikes

        return self._load_object_dict(spikes)

    def spikes_by_segment(self):
        """ Return a dictionary (indexed by Segment) of lists of
        :class:`neo.core.Spike` objects.
        """
        spikes = OrderedDict()
        units = self.units()
        for s in self.segments():
            sp = [t for t in s.spikes if t.unit in units or
                  t.unit is None]
            if sp:
                spikes[s] = sp

        nonespikes = []
        for u in self.units():
            nonespikes.extend([t for t in u.spikes if t.segment is None])
        if nonespikes:
            spikes[self.no_segment] = nonespikes

        return self._load_object_dict(spikes)

    def spikes_by_unit_and_segment(self):
        """ Return a dictionary (indexed by Unit) of dictionaries
        (indexed by Segment) of :class:`neo.core.Spike` lists.
        """
        spikes = OrderedDict()
        segments = self.segments()
        for u in self.units():
            for s in segments:
                segtrains = [t for t in u.spikes if t.segment == s]
                if segtrains:
                    if u not in spikes:
                        spikes[u] = OrderedDict()
                    spikes[u][s] = segtrains
            nonespikes = [t for t in u.spikes if t.segment is None]
            if nonespikes:
                if u not in spikes:
                    spikes[u] = OrderedDict()
                spikes[u][self.no_segment] = nonespikes

        nonespikes = OrderedDict()
        for s in self.segments():
            segspikes = [t for t in s.spikes if t.unit is None]
            if segspikes:
                nonespikes[s] = segspikes
        if nonespikes:
            spikes[self.no_unit] = nonespikes

        return self._load_object_dict(spikes)

    def events(self, include_array_events=True):
        """ Return a dictionary (indexed by Segment) of lists of
        Event objects.
        """
        ret = OrderedDict()
        for s in self.segments():
            if s.events:
                ret[s] = s.events
            if include_array_events:
                for a in s.eventarrays:
                    if s not in ret:
                        ret[s] = []
                    ret[s].extend(convert.event_array_to_events(a))
        return ret

    def labeled_events(self, label, include_array_events=True):
        """ Return a dictionary (indexed by Segment) of lists of Event
        objects with the given label.
        """
        ret = OrderedDict()
        for s in self.segments():
            events = [e for e in s.events if e.label == label]
            if events:
                ret[s] = events
            if include_array_events:
                for a in s.eventarrays:
                    if s not in ret:
                        ret[s] = []
                    events = convert.event_array_to_events(a)
                    ret[s].extend((e for e in events if e.label == label))
        return ret

    def event_arrays(self):
        """ Return a dictionary (indexed by Segment) of lists of
        EventArray objects.
        """
        ret = OrderedDict()
        for s in self.segments():
            if s.eventarrays:
                ret[s] = s.eventarrays
        return self._load_object_dict(ret)

    def epochs(self, include_array_epochs=True):
        """ Return a dictionary (indexed by Segment) of lists of
        Epoch objects.
        """
        ret = OrderedDict()
        for s in self.segments():
            if s.epochs:
                ret[s] = s.epochs
            if include_array_epochs:
                for a in s.epocharrays:
                    if s not in ret:
                        ret[s] = []
                    ret[s].extend(convert.epoch_array_to_epochs(a))
        return ret

    def labeled_epochs(self, label, include_array_epochs=True):
        """ Return a dictionary (indexed by Segment) of lists of Epoch
        objects with the given label.
        """
        ret = OrderedDict()
        for s in self.segments():
            epochs = [e for e in s.epochs if e.label == label]
            if epochs:
                ret[s] = epochs
            if include_array_epochs:
                for a in s.epocharrays:
                    if s not in ret:
                        ret[s] = []
                    epochs = convert.epoch_array_to_epochs(a)
                    ret[s].extend((e for e in epochs if e.label == label))
        return ret

    def epoch_arrays(self):
        """ Return a dictionary (indexed by Segment) of lists of
        EpochArray objects.
        """
        ret = OrderedDict()
        for s in self.segments():
            if s.epocharrays:
                ret[s] = s.epocharrays
        return self._load_object_dict(ret)

    def analog_signals(self, conversion_mode=1):
        """ Return a list of :class:`neo.core.AnalogSignal` objects.
        """
        signals = []
        channels = self.recording_channels()

        if conversion_mode == 1 or conversion_mode == 3:
            for s in self.segments():
                signals.extend([t for t in s.analogsignals
                                if t.recordingchannel in channels or
                                t.recordingchannel is None])
            for u in self.recording_channels():
                signals.extend([t for t in u.analogsignals
                                if t.segment is None])
        if conversion_mode > 1:
            for sa in self.analog_signal_arrays():
                for sig in convert.analog_signal_array_to_analog_signals(sa):
                    if (sig.recordingchannel is None or
                            sig.recordingchannel in channels):
                        signals.append(sig)

        return self._load_object_list(signals)

    def analog_signals_by_segment(self, conversion_mode=1):
        """ Return a dictionary (indexed by Segment) of lists of
        :class:`neo.core.AnalogSignal` objects.
        """
        signals = OrderedDict()
        channels = self.recording_channels()

        if conversion_mode == 1 or conversion_mode == 3:
            for s in self.segments():
                sig = []
                for c in channels:
                    sig.extend([t for t in c.analogsignals
                                if t.segment == s])
                sig.extend([t for t in s.analogsignals
                            if t.recordingchannel is None])
                if sig:
                    signals[s] = sig

            nonesignals = []
            for c in channels:
                nonesignals.extend([t for t in c.analogsignals
                                    if t.segment is None])
            if nonesignals:
                signals[self.no_segment] = nonesignals

        if conversion_mode > 1:
            for o, sa_list in \
                    self.analog_signal_arrays_by_segment().iteritems():
                for sa in sa_list:
                    for sig in \
                            convert.analog_signal_array_to_analog_signals(sa):
                        if sig.recordingchannel is None or \
                                sig.recordingchannel in channels:
                            if o not in signals:
                                signals[o] = []
                            signals[o].append(sig)

        return self._load_object_dict(signals)

    def analog_signals_by_channel(self, conversion_mode=1):
        """ Return a dictionary (indexed by RecordingChannel) of lists
        of :class:`neo.core.AnalogSignal` objects.
        """
        signals = OrderedDict()
        channels = self.recording_channels()

        if conversion_mode == 1 or conversion_mode == 3:
            segments = self.segments()
            for c in channels:
                sig = [t for t in c.analogsignals
                       if t.segment in segments or
                       t.segment is None]
                if sig:
                    signals[c] = sig

            nonesignals = []
            for s in segments:
                nonesignals.extend([t for t in s.analogsignals
                                    if t.recordingchannel is None])
            if nonesignals:
                signals[self.no_channel] = nonesignals

        if conversion_mode > 1:
            for o, sa_list in \
                    self.analog_signal_arrays_by_channelgroup().iteritems():
                for sa in sa_list:
                    for sig in \
                            convert.analog_signal_array_to_analog_signals(sa):
                        if sig.recordingchannel is None:
                            if self.no_channel not in signals:
                                signals[self.no_channel] = [sig]
                            else:
                                signals[self.no_channel].append(sig)
                        elif sig.recordingchannel in channels:
                            if sig.recordingchannel not in signals:
                                signals[sig.recordingchannel] = [sig]
                            else:
                                signals[sig.recordingchannel].append(sig)

        return self._load_object_dict(signals)

    def analog_signals_by_channel_and_segment(self, conversion_mode=1):
        """ Return a dictionary (indexed by RecordingChannel) of
        dictionaries (indexed by Segment) of :class:`neo.core.AnalogSignal`
        lists.
        """
        signals = OrderedDict()
        channels = self.recording_channels()

        if conversion_mode == 1 or conversion_mode == 3:
            segments = self.segments()
            for c in channels:
                for s in segments:
                    segsignals = [t for t in c.analogsignals if t.segment == s]
                    if segsignals:
                        if c not in signals:
                            signals[c] = OrderedDict()
                        signals[c][s] = segsignals
                nonesignals = [t for t in c.analogsignals if t.segment is None]
                if nonesignals:
                    if c not in signals:
                        signals[c] = OrderedDict()
                    signals[c][self.no_segment] = nonesignals

            nonesignals = OrderedDict()
            for s in self.segments():
                segsignals = [t for t in s.analogsignals
                              if t.recordingchannel is None]
                if segsignals:
                    nonesignals[s] = segsignals
            if nonesignals:
                signals[self.no_channel] = nonesignals

        if conversion_mode > 1:
            sigs = self.analog_signal_arrays_by_channelgroup_and_segment()
            for cg, inner in sigs.iteritems():
                for seg, sa_list in inner.iteritems():
                    for sa in sa_list:
                        for sig in convert.analog_signal_array_to_analog_signals(sa):
                            chan = sig.recordingchannel
                            if chan not in channels:
                                continue
                            if chan not in signals:
                                signals[chan] = OrderedDict()
                            if seg not in signals[chan]:
                                signals[chan][seg] = []
                            signals[chan][seg].append(sig)

        return self._load_object_dict(signals)

    def analog_signal_arrays(self):
        """ Return a list of :class:`neo.core.AnalogSignalArray` objects.
        """
        signals = []
        channelgroups = self.recording_channel_groups()
        for s in self.segments():
            signals.extend([t for t in s.analogsignalarrays
                            if t.recordingchannelgroup in channelgroups or
                            t.recordingchannelgroup is None])
        for u in channelgroups:
            signals.extend([t for t in u.analogsignalarrays
                            if t.segment is None])

        return self._load_object_list(signals)

    def analog_signal_arrays_by_segment(self):
        """ Return a dictionary (indexed by Segment) of lists of
        :class:`neo.core.AnalogSignalArray` objects.
        """
        signals = OrderedDict()
        channelgroups = self.recording_channel_groups()
        for s in self.segments():
            sa = []
            for c in channelgroups:
                sa.extend([t for t in c.analogsignalarrays
                           if t.segment == s])
            sa.extend([t for t in s.analogsignalarrays
                       if t.recordingchannelgroup is None])
            if sa:
                signals[s] = sa

        nonesignals = []
        for c in channelgroups:
            nonesignals.extend([t for t in c.analogsignalarrays
                                if t.segment is None])
        if nonesignals:
            signals[self.no_segment] = nonesignals

        return self._load_object_dict(signals)

    def analog_signal_arrays_by_channelgroup(self):
        """ Return a dictionary (indexed by RecordingChannelGroup) of
        lists of :class:`neo.core.AnalogSignalArray` objects.
        """
        signals = OrderedDict()
        segments = self.segments()
        for c in self.recording_channel_groups():
            sa = [t for t in c.analogsignalarrays
                  if t.segment in segments]
            if sa:
                signals[c] = sa

        nonesignals = []
        for s in segments:
            nonesignals.extend([t for t in s.analogsignalarrays
                                if t.recordingchannelgroup is None])
        if nonesignals:
            signals[self.no_channelgroup] = nonesignals

        return self._load_object_dict(signals)

    def analog_signal_arrays_by_channelgroup_and_segment(self):
        """ Return a dictionary (indexed by RecordingChannelGroup) of
        dictionaries (indexed by Segment) of
        :class:`neo.core.AnalogSignalArray` lists.
        """
        signals = OrderedDict()
        segments = self.segments()
        for c in self.recording_channel_groups():
            for s in segments:
                segsignals = [t for t in c.analogsignalarrays
                              if t.segment == s]
                if segsignals:
                    if c not in signals:
                        signals[c] = OrderedDict()
                    signals[c][s] = segsignals
            nonesignals = [t for t in c.analogsignalarrays
                           if t.segment is None]
            if nonesignals:
                if c not in signals:
                    signals[c] = OrderedDict()
                signals[c][self.no_segment] = nonesignals

        nonesignals = OrderedDict()
        for s in self.segments():
            segsignals = [t for t in s.analogsignalarrays
                          if t.recordingchannelgroup is None]
            if segsignals:
                nonesignals[s] = segsignals
        if nonesignals:
            signals[self.no_channelgroup] = nonesignals

        return self._load_object_dict(signals)


atexit.register(NeoDataProvider.clear)


""" This module gives access to all members of
:mod:`guidata.dataset.dataitems` and :mod:`guidata.dataset.datatypes`.
If :mod:`guidata` cannot be imported, the module offers suitable dummy
objects instead (e.g. for use on a server).
"""
try:
    from guidata.dataset.dataitems import *
    from guidata.dataset.datatypes import *
except ImportError, e:
    import datetime
    import scipy as sp

    # datatypes dummies
    class DataItem(object):
        def __init__(self, *args, **kwargs):
            self._default = None
            if 'default' in kwargs:
                self._default = kwargs['default']

        def get_prop(self, realm, name, default=None):
            pass

        def get_prop_value(self, realm, instance, name, default=None):
            pass

        def set_prop(self, realm, **kwargs):
            return self

        def set_pos(self, col=0, colspan=None):
            return self

        def get_help(self, instance):
            pass

        def get_auto_help(self, instance):
            pass

        def format_string(self, instance, value, fmt, func):
            pass

        def get_string_value(self, instance):
            pass

        def set_name(self, new_name):
            self._name = new_name

        def set_from_string(self, instance, string_value):
            pass

        def set_default(self, instance):
            self.__set__(instance, self._default)

        def accept(self, visitor):
            pass

        def __set__(self, instance, value):
            setattr(instance, "_"+self._name, value)

        def __get__(self, instance, klass):
            if instance is not None:
                return getattr(instance, "_"+self._name, self._default)
            else:
                return self

        def get_value(self, instance):
            return self.__get__(instance, instance.__class__)

        def check_item(self, instance):
            pass

        def check_value(self, instance, value):
            pass

        def from_string(self, instance, string_value):
            pass

        def bind(self, instance):
            return DataItemVariable(self, instance)

        def serialize(self, instance, writer):
            pass

        def deserialize(self, instance, reader):
            pass

    class DataSetMeta(type):
        """ DataSet metaclass.

        Create class attribute `_items`: list of the DataSet class attributes.
        Also make sure that all data items have the correct `_name` attribute.
        """
        def __new__(mcs, name, bases, dct):
            items = []
            for base in bases:
                if getattr(base, "__metaclass__", None) is DataSetMeta:
                    for item in base._items:
                        items.append(item)

            for attrname, value in dct.items():
                if isinstance(value, DataItem):
                    value.set_name(attrname)
                    items.append(value)
            dct["_items"] = items
            return type.__new__(mcs, name, bases, dct)

    class DataSet(object):
        __metaclass__ = DataSetMeta

        def __init__(self, title=None, comment=None, icon=''):
            self.set_defaults()

        def _get_translation(self):
            pass

        def _compute_title_and_comment(self):
            pass

        def get_title(self):
            pass

        def get_comment(self):
            pass

        def get_icon(self):
            pass

        def set_defaults(self):
            for item in self._items:
                item.set_default(self)

        def check(self):
            pass

        def text_edit(self):
            pass

        def edit(self, parent=None, apply=None):
            pass

        def view(self, parent=None):
            pass

        def to_string(self, debug=False, indent=None, align=False):
            pass

        def accept(self, vis):
            pass

        def serialize(self, writer):
            pass

        def deserialize(self, reader):
            pass

        def read_config(self, conf, section, option):
            pass

        def write_config(self, conf, section, option):
            pass

        @classmethod
        def set_global_prop(cls, realm, **kwargs):
            pass

    class ItemProperty(object):
        def __init__(self, callable=None):
            pass

        def __call__(self, instance, item, value):
            pass

        def set(self, instance, item, value):
            pass

    class FormatProp(ItemProperty):
        def __init__(self, fmt, ignore_error=True):
            pass

    class GetAttrProp(ItemProperty):
        pass

    class ValueProp(ItemProperty):
        pass

    class NotProp(ItemProperty):
        pass

    class DataItemVariable(object):
        def __init__(self, item, instance):
            self.item = item
            self.instance = instance

        def get_prop_value(self, realm, name, default=None):
            pass

        def get_prop(self, realm, name, default=None):
            pass

        def get_help(self):
            pass

        def get_auto_help(self):
            pass

        def get_string_value(self):
            pass

        def set_default(self):
            return self.item.set_default(self.instance)

        def get(self):
            return self.item.get_value(self.instance)

        def set(self, value):
            return self.item.__set__(self.instance, value)

        def set_from_string(self, string_value):
            pass

        def check_item(self):
            pass

        def check_value(self, value):
            pass

        def from_string(self, string_value):
            pass

        def label(self):
            pass


    class GroupItem(DataItem):
        pass

    class BeginGroup(DataItem):
        pass

    class EndGroup(DataItem):
        pass

    class TabGroupItem(GroupItem):
        pass

    class BeginTabGroup(BeginGroup):
        def get_group(self):
            pass

    class EndTabGroup(EndGroup):
        pass


    # dataitems dummies
    class NumericTypeItem(DataItem):
        def __init__(self, *args, **kwargs):
            super(NumericTypeItem, self).__init__(*args, **kwargs)

    class FloatItem(NumericTypeItem):
        def __init__(self, *args, **kwargs):
            super(FloatItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = float()

    class IntItem(NumericTypeItem):
        def __init__(self, *args, **kwargs):
            super(IntItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = int()

    class StringItem(DataItem):
        def __init__(self, *args, **kwargs):
            super(StringItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = str()

    class BoolItem(DataItem):
        def __init__(self, *args, **kwargs):
            super(BoolItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = bool()

    class DateItem(DataItem):
        def __init__(self, *args, **kwargs):
            super(DateItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = datetime.date.today()

    class DateTimeItem(DateItem):
        def __init__(self, *args, **kwargs):
            super(DateTimeItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = datetime.datetime.now()

    class ColorItem(StringItem):
        def __init__(self, *args, **kwargs):
            super(ColorItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = str()

    class FileSaveItem(StringItem):
        def __init__(self, *args, **kwargs):
            super(FileSaveItem, self).__init__(*args, **kwargs)

    class FileOpenItem(FileSaveItem):
        def __init__(self, *args, **kwargs):
            super(FileOpenItem, self).__init__(*args, **kwargs)

    class FilesOpenItem(FileSaveItem):
        def __init__(self, *args, **kwargs):
            super(FilesOpenItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = list()

    class DirectoryItem(StringItem):
        def __init__(self, *args, **kwargs):
            super(DirectoryItem, self).__init__(*args, **kwargs)

    class ChoiceItem(DataItem):
        def __init__(self, *args, **kwargs):
            super(ChoiceItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = int()

    class MultipleChoiceItem(ChoiceItem):
        def __init__(self, *args, **kwargs):
            super(MultipleChoiceItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = list()

    class ImageChoiceItem(ChoiceItem):
        def __init__(self, *args, **kwargs):
            super(ImageChoiceItem, self).__init__(*args, **kwargs)

    class FloatArrayItem(DataItem):
        def __init__(self, *args, **kwargs):
            super(FloatArrayItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = sp.array([])

    class ButtonItem(DataItem):
        def __init__(self, *args, **kwargs):
            super(ButtonItem, self).__init__(*args, **kwargs)

    class DictItem(ButtonItem):
        def __init__(self, *args, **kwargs):
            super(DictItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = dict()

    class FontFamilyItem(StringItem):
        def __init__(self, *args, **kwargs):
            super(FontFamilyItem, self).__init__(*args, **kwargs)
            if self._default is None:
                self._default = str()


import scipy as sp
import quantities as pq

from guiqwt.plot import BaseImageWidget
from guiqwt.builder import make

from ..progress_indicator import ProgressIndicator
from .. import SpykeException
from ..stationarity import spike_amplitude_histogram as sah
import helper
from dialog import PlotDialog


@helper.needs_qt
def spike_amplitude_histogram(trains, num_bins, uniform_y_scale=True,
                              x_unit=pq.uV, progress=None):
    """ Create a spike amplitude histogram.

    This plot is useful to assess the drift in spike amplitude over a longer
    recording. It shows histograms (one for each `trains` entry, e.g. segment)
    of maximum and minimum spike amplitudes.

    :param list trains: A list of lists of :class:`neo.core.SpikeTrain`
        objects. Each entry of the outer list will be one point on the
        x-axis (they could correspond to segments), all amplitude occurences
        of spikes contained in the inner list will be added up.
    :param int num_bins: Number of bins for the histograms.
    :param bool uniform_y_scale: If True, the histogram for each channel
        will use the same bins. Otherwise, the minimum bin range is computed
        separately for each channel.
    :param Quantity x_unit: Unit of Y-Axis.
    :param progress: Set this parameter to report progress.
    :type progress: :class:`spykeutils.progress_indicator.ProgressIndicator`
    :return:
    """
    if not trains:
        raise SpykeException('No spikes trains for Spike Amplitude Histogram!')
    if not progress:
        progress = ProgressIndicator()

    hist, down, up = sah(trains, num_bins, uniform_y_scale, x_unit, progress)
    num_channels = len(down)

    columns = int(round(sp.sqrt(num_channels)))

    win = PlotDialog(toolbar=True, wintitle="Spike Amplitude Histogram")
    for c in xrange(num_channels):
        pW = BaseImageWidget(
            win, yreverse=False, lock_aspect_ratio=False)
        plot = pW.plot
        img = make.image(sp.log(hist[:, :, c] + 1),
                         ydata=[down[c], up[c]],
                         interpolation='nearest')
        plot.add_item(img)
        plot.set_axis_title(plot.Y_LEFT, 'Amplitude')
        plot.set_axis_unit(plot.Y_LEFT, x_unit.dimensionality.string)
        win.add_plot_widget(pW, c, column=c % columns)

    progress.done()
    win.add_custom_image_tools()
    win.add_x_synchronization_option(True, range(num_channels))
    win.add_y_synchronization_option(uniform_y_scale,
        range(num_channels))
    win.show()

    return win


from __future__ import absolute_import
import quantities as pq


# At least up to quantities 0.10.1 the additional arguments to the min and max
# function did not get passed along.
# A fix already exists:
# <https://github.com/dopplershift/python-quantities/commit/30e8812ac15f551c65311d808c2a004f53913a25>
# Also a pull request exists which has not been merged at the time of writing
# 01/18/2013:
# <https://github.com/python-quantities/python-quantities/pull/41>

def _Quanitity_max(self, axis=None, out=None):
    return pq.Quantity(
        self.magnitude.max(axis, out),
        self.dimensionality,
        copy=False
    )
pq.Quantity.max = _Quanitity_max


def _Quanitity_min(self, axis=None, out=None):
    return pq.Quantity(
        self.magnitude.min(axis, out),
        self.dimensionality,
        copy=False
    )
pq.Quantity.min = _Quanitity_min
